"""
This module contains the class definitions for different laser pulses.

The implementation of the Laguerre-Gauss and Flattened Gaussian
profiles, as well as the SummedPulse approach is based on FBPIC
(https://github.com/fbpic/fbpic).

Authors: Angel Ferran Pousa, Remi Lehe, Manuel Kirchen, Pierre Pelletier.

"""

from typing import Optional, Union, Iterable

import numpy as np
import scipy.constants as ct
from scipy.special import genlaguerre, binom
from scipy.ndimage import gaussian_filter

try:
    from lasy.profiles.from_openpmd_profile import FromOpenPMDProfile
    from lasy.laser import Laser
    from lasy.utils.laser_utils import field_to_vector_potential

    lasy_installed = True
except ImportError:
    lasy_installed = False


from .envelope_solver import evolve_envelope
from .envelope_solver_non_centered import evolve_envelope_non_centered
from wake_t.fields.interpolation import interpolate_rz_field


class LaserPulse:
    """Base class for all Laser pulses.

    Parameters
    ----------
    l_0 : float
        Laser wavelength in units of m.
    polarization : str, optional
        Polarization of the laser pulse. Accepted values are 'linear'
        (default) or 'circular'.

    """

    def __init__(self, l_0: float, polarization: Optional[str] = "linear") -> None:
        self.l_0 = l_0
        self.polarization = polarization
        self.a_env = None
        self.solver_params = None
        self.n_steps = 0

    def __add__(self, pulse_2):
        """Overload the add operator to allow summing of laser pulses."""
        return SummedPulse(self, pulse_2)

    def set_envelope_solver_params(
        self,
        xi_min: float,
        xi_max: float,
        r_max: float,
        nz: int,
        nr: int,
        dt: float,
        nt: Optional[int] = 1,
        subgrid_nz: Optional[int] = None,
        subgrid_nr: Optional[int] = None,
        use_phase: Optional[bool] = True,
    ) -> None:
        """
        Set the parameters for the laser envelope solver.

        Parameters
        ----------
        xi_min, xi_max : float
            Position of the left and right boundaries of the the grid in SI
            units.
        r_max : float
            Radial extension of the grid in SI units.
        nz, nr : int
            Number of grid points along the longitudinal and radial directions.
        dt : float
            Time step, in SI units, that the laser pulse should advance every
            time `evolve` is called.
        nt : int
            Number of sub-time-steps that should be computed every time
            `evolve` is called. The internal time step used by the solver is
            therefore dt/nt, so that the laser effectively advances by `dt`
            every time `evolve` is called. All these time steps are therefore
            computed using the same `chi`.
        subgrid_nz, subgrid_nr : int, optional
            If specified, run laser envelope solver in a subgrid of resolution
            (subgrid_nz, subgrid_nr), which is different than the size of the
            plasma grid (nz, nr). Linear interpolation is used to transform
            the plasma susceptibility into the laser envelope grid, and to
            transform the laser envelope into the plasma grid.
        use_phase : bool
            Determines whether to take into account the terms related to the
            longitudinal derivative of the complex phase in the envelope
            solver.

        """
        if nt < 1:
            raise ValueError(
                "Number of laser envelope substeps cannot be smaller than 1."
            )

        # Determine whether to run laser envelope in a subgrid.
        self.use_subgrid = subgrid_nz is not None or subgrid_nr is not None
        if self.use_subgrid:
            subgrid_nz = nz if subgrid_nz is None else subgrid_nz
            subgrid_nr = nr if subgrid_nr is None else subgrid_nr
            self._create_laser_subgrid(
                nz, nr, subgrid_nz, subgrid_nr, xi_max, xi_min, r_max
            )

        solver_params = {
            "zmin": xi_min,
            "zmax": xi_max,
            "rmax": r_max,
            "nz": subgrid_nz if self.use_subgrid else nz,
            "nr": subgrid_nr if self.use_subgrid else nr,
            "nt": nt,
            "dt": dt / nt,
            "use_phase": use_phase,
        }
        # Check if laser pulse has already been initialized.
        if self.n_steps > 0:
            # Check that grid parameters have not changed.
            if (
                solver_params["zmin"] != self.solver_params["zmin"]
                and solver_params["zmax"] != self.solver_params["zmax"]
                and solver_params["rmax"] != self.solver_params["rmax"]
                and solver_params["nz"] != self.solver_params["nz"]
                and solver_params["nr"] != self.solver_params["nr"]
            ):
                raise ValueError(
                    "Laser envelope grid parameters cannot be changed once "
                    "envelope has been initialized."
                )
            # If time step has changed, set `n_steps` to 0 so that the
            # non-centered envelope solver is used for the next step.
            if solver_params["dt"] != self.solver_params["dt"]:
                self.n_steps = 0
        self.solver_params = solver_params

    def initialize_envelope(self) -> None:
        """Initialize laser envelope arrays."""
        if self.solver_params is None:
            raise ValueError(
                "Envelope solver parameters not yet set.Cannot initialize envelope."
            )
        if self.a_env is None:
            z_min = self.solver_params["zmin"]
            z_max = self.solver_params["zmax"]
            r_max = self.solver_params["rmax"]
            nz = self.solver_params["nz"]
            nr = self.solver_params["nr"]
            dr = r_max / nr
            z = np.linspace(z_min, z_max, nz)
            r = np.linspace(dr / 2, r_max - dr / 2, nr)
            ZZ, RR = np.meshgrid(z, r, indexing="ij")
            self._a_env_old = np.zeros((nz + 2, nr), dtype=np.complex128)
            self._a_env = np.zeros((nz + 2, nr), dtype=np.complex128)
            self._a_env[0:-2] = self.envelope_function(ZZ, RR, 0.0)
            self._update_output_envelope()

    def get_envelope(self) -> np.ndarray:
        """Get the current laser envelope array."""
        return self.a_env

    def evolve(self, chi: np.ndarray, n_p: float) -> None:
        """
        Evolve laser envelope to next time step.

        Parameters
        ----------
        chi : ndarray
            A (nz x nr) array containing the plasma susceptibility.
        n_p : float
            Plasma density in SI units.
        """
        k_0 = 2 * np.pi / self.l_0
        k_p = np.sqrt(ct.e**2 * n_p / (ct.m_e * ct.epsilon_0)) / ct.c

        # If needed, interpolate chi to subgrid.
        if self.use_subgrid:
            chi = self._interpolate_chi_to_subgrid(chi)

        # Compute evolution.
        if self.n_steps == 0:
            evolve_envelope_non_centered(
                self._a_env, self._a_env_old, chi, k_0, k_p, **self.solver_params
            )
        else:
            evolve_envelope(
                self._a_env, self._a_env_old, chi, k_0, k_p, **self.solver_params
            )

        # Update arrays and step count.
        self._update_output_envelope()
        self.n_steps += 1

    def get_group_velocity(self, n_p: float) -> float:
        """
        Get group velocity of the laser pulse for a given plasma density.

        Parameters
        ----------
        n_p : float
            Plasma density in units of m^{-3}.

        Returns
        -------
        A float containing the group velocity.
        """
        w_p = np.sqrt(n_p * ct.e**2 / (ct.m_e * ct.epsilon_0))
        k = 2 * np.pi / self.l_0
        v_g = k * ct.c**2 / np.sqrt(w_p**2 + k**2 * ct.c**2) / ct.c
        return v_g

    def envelope_function(
        self, xi: np.ndarray, r: np.ndarray, z_pos: float
    ) -> np.ndarray:
        """Return the complex envelope of the laser pulse."""
        return self._envelope_function(xi, r, z_pos)

    def _envelope_function(self, xi, r, z_pos):
        return np.zeros_like(r)

    def _create_laser_subgrid(
        self, nz, nr, subgrid_nz, subgrid_nr, xi_max, xi_min, r_max
    ):
        """
        Create the parameters needed to run the laser envelope in a subgrid.
        """
        # Grid spacing and minimum radius of the main grid.
        dr = r_max / nr
        dz = (xi_max - xi_min) / (nz - 1)
        grid_r_min = dr / 2
        grid_r_max = r_max - dr / 2

        # Grid spacing and minimum radius of the subgrid.
        subgrid_dr = r_max / subgrid_nr
        subgrid_dz = (xi_max - xi_min) / (subgrid_nz - 1)
        subgrid_r_min = subgrid_dr / 2
        subgrid_r_max = r_max - subgrid_dr / 2

        # Store parameters in dictionary.
        self.subgrid_params = {
            "grid": {
                "nz": nz,
                "nr": nr,
                "z_min": xi_min,
                "r_min": grid_r_min,
                "dr": dr,
                "dz": dz,
                "z": np.linspace(xi_min, xi_max, nz),
                "r": np.linspace(grid_r_min, grid_r_max, nr),
            },
            "subgrid": {
                "nz": subgrid_nz,
                "nr": subgrid_nr,
                "z_min": xi_min,
                "r_min": subgrid_r_min,
                "dr": subgrid_dr,
                "dz": subgrid_dz,
                "z": np.linspace(xi_min, xi_max, subgrid_nz),
                "r": np.linspace(subgrid_r_min, subgrid_r_max, subgrid_nr),
                "chi": np.zeros((subgrid_nz, subgrid_nr)),
            },
        }

    def _update_output_envelope(self):
        """Update the publicly-accessible laser envelope array."""
        # If running on a subgrid, interpolate envelope array to main grid.
        if self.use_subgrid:
            if self.a_env is None:
                nz = self.subgrid_params["grid"]["nz"]
                nr = self.subgrid_params["grid"]["nr"]
                self.a_env = np.zeros((nz, nr), dtype=np.complex128)
            z_min = self.subgrid_params["subgrid"]["z_min"]
            r_min = self.subgrid_params["subgrid"]["r_min"]
            dz = self.subgrid_params["subgrid"]["dz"]
            dr = self.subgrid_params["subgrid"]["dr"]
            z_f = self.subgrid_params["grid"]["z"]
            r_f = self.subgrid_params["grid"]["r"]
            interpolate_rz_field(
                self._a_env[0:-2], z_min, r_min, dz, dr, z_f, r_f, self.a_env
            )
        # Otherwise, simply remove guard cells.
        else:
            if self.a_env is None:
                nz = self.solver_params["nz"]
                nr = self.solver_params["nr"]
                self.a_env = np.zeros((nz, nr), dtype=np.complex128)
            self.a_env[:] = self._a_env[0:-2]

    def _interpolate_chi_to_subgrid(self, chi):
        """Interpolate the plasma susceptibility to the envelope subgrid."""
        z_min = self.subgrid_params["grid"]["z_min"]
        r_min = self.subgrid_params["grid"]["r_min"]
        dz = self.subgrid_params["grid"]["dz"]
        dr = self.subgrid_params["grid"]["dr"]
        subgrid_chi = self.subgrid_params["subgrid"]["chi"]
        z_f = self.subgrid_params["subgrid"]["z"]
        r_f = self.subgrid_params["subgrid"]["r"]
        interpolate_rz_field(chi, z_min, r_min, dz, dr, z_f, r_f, subgrid_chi)
        return subgrid_chi


class SummedPulse(LaserPulse):
    """Class defining a laser pulse made up of the addition of two pulses.

    Parameters
    ----------
    pulse_1 : LaserPulse
        A LaserPulse instance.
    pulse_2 : LaserPulse
        Another LaserPulse instance to be summed to `pulse_1`.
    """

    def __init__(self, pulse_1: LaserPulse, pulse_2: LaserPulse) -> None:
        if pulse_1.l_0 != pulse_2.l_0:
            raise ValueError(
                "Only laser pulses with the same central wavelength"
                " can currently be summed."
            )
        super().__init__(pulse_1.l_0)
        self.pulse_1 = pulse_1
        self.pulse_2 = pulse_2

    def _envelope_function(self, xi, r, z_pos):
        """Return the summed envelope of the two pulses."""
        a_env_1 = self.pulse_1.envelope_function(xi, r, z_pos)
        a_env_2 = self.pulse_2.envelope_function(xi, r, z_pos)
        return a_env_1 + a_env_2


class GaussianPulse(LaserPulse):
    """Class defining a Gaussian laser pulse.

    Parameters
    ----------
    xi_c : float
        Initial central position of the pulse along xi in units of m.
    a_0 : float
        The peak normalized vector potential at the focal plane.
    w_0 : float
        Spot size of the laser pulse, in units of m, at the focal plane.
    tau : float
        Longitudinal pulse length (FWHM in intensity) in units of s.
    z_foc : float, optional
        Focal position of the pulse.
    l_0 : float, optional
        Laser wavelength in units of m. By default, a Ti:Sa laser with
        `l_0=0.8e-6` is assumed.
    cep_phase : float, optional
        The Carrier Envelope Phase (CEP) in radian. This is the phase of
        the laser oscillation at the position where the envelope is
        maximum.
    polarization : str, optional
        Polarization of the laser pulse. Accepted values are 'linear'
        (default) or 'circular'.
    """

    def __init__(
        self,
        xi_c: float,
        a_0: float,
        w_0: float,
        tau: float,
        z_foc: Optional[float] = 0.0,
        l_0: Optional[float] = 0.8e-6,
        cep_phase: Optional[float] = 0.0,
        polarization: Optional[str] = "linear",
    ) -> None:
        super().__init__(l_0=l_0, polarization=polarization)
        self.xi_c = xi_c
        self.a_0 = a_0
        self.tau = tau
        self.w_0 = w_0
        self.z_foc = z_foc
        self.z_r = np.pi * w_0**2 / l_0
        self.cep_phase = cep_phase

    def _envelope_function(self, xi, r, z_pos):
        """
        Complex envelope of a Gaussian beam in the paraxial approximation.
        """
        z = xi - self.xi_c + z_pos
        diff_factor = 1.0 + 1j * (z - self.z_foc) / self.z_r
        s_z = self.tau * ct.c / (2 * np.sqrt(2 * np.log(2))) * np.sqrt(2)
        # Phases
        exp_cep = -1j * self.cep_phase
        exp_r = -(r**2) / (self.w_0**2 * diff_factor)
        exp_z = -((xi - self.xi_c) ** 2) / (2 * s_z**2)
        # Profile
        gaussian_profile = np.exp(exp_cep + exp_r + exp_z)
        # Amplitude
        avg_amplitude = self.a_0
        return avg_amplitude / diff_factor * gaussian_profile


class LaguerreGaussPulse(LaserPulse):
    """Class defining a Laguerre-Gauss pulse.

    Due to the cylindrical geometry of Wake-T, only the `0` azimuthal mode
    is supported.

    Parameters
    ----------
    xi_c : float
        Initial central position of the pulse along xi in units of m.
    p : int (positive)
        The order of the Laguerre polynomial. Increasing ``p`` increases
        the number of "rings" in the radial intensity profile of the laser.
    a_0 : float
        The peak normalized vector potential at the focal plane, defined
        so that the total energy of the pulse is the same as that of a
        Gaussian pulse with the same `a_0`, `w_0` `tau`.
        (i.e. The energy of the pulse is independent of `p`.)
    w_0 : float
        Spot size of the laser pulse, in units of m, at the focal plane.
    tau : float
        Longitudinal pulse length (FWHM in intensity) in units of s.
    z_foc : float, optional
        Focal position of the pulse.
    l_0 : float, optional
        Laser wavelength in units of m. By default, a Ti:Sa laser with
        `l_0=0.8e-6` is assumed.
    cep_phase : float, optional
        The Carrier Envelope Phase (CEP) in radian. This is the phase of
        the laser oscillation at the position where the envelope is
        maximum.
    polarization : str, optional
        Polarization of the laser pulse. Accepted values are 'linear'
        (default) or 'circular'.

    """

    def __init__(
        self,
        xi_c: float,
        p: int,
        a_0: float,
        w_0: float,
        tau: float,
        z_foc: Optional[float] = 0.0,
        l_0: Optional[float] = 0.8e-6,
        cep_phase: Optional[float] = 0.0,
        polarization: Optional[str] = "linear",
    ) -> None:
        # Initialize parent class
        super().__init__(l_0=l_0, polarization=polarization)

        # If no focal plane position is given, use xi_c
        if z_foc is None:
            z_foc = xi_c

        # Store the parameters
        self.p = p
        self.laguerre_pm = genlaguerre(self.p, 0)  # Laguerre polynomial
        self.z_r = np.pi * w_0**2 / l_0
        self.z_foc = z_foc
        self.xi_c = xi_c
        self.a0 = a_0
        self.w0 = w_0
        self.cep_phase = cep_phase
        self.tau = tau

    def _envelope_function(self, xi, r, z_pos):
        """Complex envelope of a Laguerre-Gauss beam."""
        z = xi - self.xi_c + z_pos
        # Diffraction factor, waist and Gouy phase
        diffract_factor = 1.0 + 1j * (z - self.z_foc) / self.z_r
        s_z = self.tau * ct.c / (2 * np.sqrt(2 * np.log(2))) * np.sqrt(2)
        w = self.w0 * abs(diffract_factor)
        psi = np.angle(diffract_factor)
        # Calculate the scaled radius
        scaled_radius_squared = 2 * r**2 / w**2
        # Calculate the argument of the complex exponential
        exp_argument = (
            -1j * self.cep_phase
            - r**2 / (self.w0**2 * diffract_factor)
            - (xi - self.xi_c) ** 2 / (2 * s_z**2)
            - 1j * (2 * self.p) * psi
        )  # *Additional* Gouy phase
        # Get the transverse profile
        profile = (
            np.exp(exp_argument)
            / diffract_factor
            * self.laguerre_pm(scaled_radius_squared)
        )

        a = self.a0 * profile
        return a


class FlattenedGaussianPulse(LaserPulse):
    """Class defining a flattened Gaussian pulse.

    The laser pulse is defined such that the transverse intensity
    profile is a flattened Gaussian far from focus, and a distribution
    with rings in the focal plane. (See `Santarsiero et al., J.
    Modern Optics, 1997 <http://doi.org/10.1080/09500349708232927>`_)
    Increasing the parameter ``N`` increases the
    flatness of the transverse profile far from focus,
    and increases the number of rings in the focal plane.

    Parameters
    ----------
    xi_c : float
        Initial central position of the pulse along xi in units of m.
    a_0: float
        The peak normalized vector potential at the focal plane.
    w_0 : float
        Spot size of the laser pulse, in units of m, at the focal plane.
    tau : float
        Longitudinal pulse length (FWHM in intensity) in units of s.
    N : int, optional
        Determines the "flatness" of the transverse profile, far from
        focus.
        Default: ``N=6`` ; somewhat close to an 8th order supergaussian.
    z_foc : float, optional
        Focal position of the pulse.
    l_0 : float, optional
        Laser wavelength in units of m. By default, a Ti:Sa laser with
        `l_0=0.8e-6` is assumed.
    cep_phase : float, optional
        The Carrier Envelope Phase (CEP) in radian. This is the phase of
        the laser oscillation at the position where the envelope is
        maximum.
    polarization : str, optional
        Polarization of the laser pulse. Accepted values are 'linear'
        (default) or 'circular'.

    """

    def __init__(
        self,
        xi_c: float,
        a_0: float,
        w_0: float,
        tau: float,
        N: Optional[int] = 6,
        z_foc: Optional[float] = 0.0,
        l_0: Optional[float] = 0.8e-6,
        cep_phase: Optional[float] = 0.0,
        polarization: Optional[str] = "linear",
    ) -> None:
        # Initialize parent class.
        super().__init__(l_0=l_0, polarization=polarization)

        # Store parameters.
        self.xi_c = xi_c

        # Ensure that N is an integer.
        N = int(round(N))
        # Calculate effective waist of the Laguerre-Gauss modes, at focus.
        w_foc = w_0 * (N + 1) ** 0.5

        # Sum the Laguerre-Gauss modes that constitute this pulse.
        # See equation (2) and (3) in Santarsiero et al.
        for n in range(N + 1):
            cep_phase_n = cep_phase + 2 * n * np.pi / 2
            m_values = np.arange(n, N + 1)
            cn = (-1) ** n * np.sum(0.5**m_values * binom(m_values, n)) / (N + 1)
            pulse = LaguerreGaussPulse(
                xi_c=xi_c,
                p=n,
                a_0=cn * a_0,
                cep_phase=cep_phase_n,
                w_0=w_foc,
                tau=tau,
                z_foc=z_foc,
                l_0=l_0,
                polarization=polarization,
            )
            if n == 0:
                summed_pulse = pulse
            else:
                summed_pulse += pulse

        # Register the summed_pulse.
        self.summed_pulse = summed_pulse

    def _envelope_function(self, xi, r, z_pos):
        """Complex envelope of the flattened Gaussian beam."""
        return self.summed_pulse.envelope_function(xi, r, z_pos)


class OpenPMDPulse(LaserPulse):
    """Read a laser pulse from an openPMD file.

    This class requires `lasy <https://lasydoc.readthedocs.io>`_ to be
    installed.

    Parameters
    ----------
    file_name : string
        Name of openPMD file, including path, to read the laser field or envelope from.
        Either specify the exact file name (e.g. ``file_name="/path/data_00001.h5"``)
        or a file pattern + iteration (e.g. ``file_name="/path/data%T.h5", iteration=1``).
    envelope_name : string (optional)
        The name of the envelope field (this is not prescribed by the openPMD standard for the envelope).
        If specified, an envelope field is expected from the openPMD file. Otherwise, a full electric field is assumed.
    iteration : int (optional)
        The iteration to read from the openPMD file. If not specified, the last iteration is used.
    t_start : float, optional
        The initialization of this class aligns the right (spatial) edge
        of the Wake-T grid with the left (temporal) edge of the Lasy grid,
        regardless of the actual time values in the lasy file. The `t_start`
        parameter introduces a time delay to the initialized laser, allowing
        for precise adjustment of the pulse position in the Wake-T grid.
    smooth_edges : bool, optional
        Whether to smooth the edges of the laser profile along `r` using a
        super-Gaussian function of power 8. This is useful when the laser
        profile in the openPMD file has a sharp edge (e.g., due to the finite
        width of the domain in `r`). Smoothing this edge can help reduce noise
        in the simulation. By default `False`.
    apply_gaussian_filter : bool, optional
        Whether to apply a Gaussian filter to the laser profile. This is
        useful, for example, when the openpmd laser pulse comes from a
        noisy simulation. In this case, applying the filter can improve the
        stability of the simulation. By default `False`.
    gaussian_filter_sigma : scalar or sequence of scalars
        Standard deviation for Gaussian kernel. The standard
        deviations of the Gaussian filter are given for each axis as a
        sequence, or as a single number, in which case it is equal for
        all axes. By default `(5, 0)`, which only smooths along the radial
        direction.

    Notes
    -----
    When the grid of the openPMD laser pulse and the grid of the Wake-T
    simulation have a different extent or resolution, the original
    pulse will be linearly interpolated into the Wake-T grid. This is can
    sometimes lead to numerical issues that manifest as a radial oscillation
    of the laser pulse (and, as a result, of the plasma wake). To mitigate
    this, it is important to avoid interpolating along the radial direction.
    This can be achieved if the ratios between the original and the Wake-T
    resolution and extent are an integer.
    """

    def __init__(
        self,
        file_name: str,
        envelope_name: Optional[str] = None,
        iteration: Optional[int] = None,
        t_start: Optional[float] = 0.0,
        smooth_edges: Optional[bool] = False,
        apply_gaussian_filter: Optional[bool] = False,
        gaussian_filter_sigma: Optional[Union[int, float, Iterable]] = (5, 0),
    ) -> None:
        assert lasy_installed, (
            "Using an `OpenPMDPulse` requires `lasy` to be installed. "
            "You can do so with `pip install lasy`."
        )
        self.lasy_profile = FromOpenPMDProfile(
            file_name=file_name,
            envelope_name=envelope_name,
            iteration=iteration,
        )
        pol = self.lasy_profile.pol
        assert np.isclose(np.abs(pol[0]) ** 2 + np.abs(pol[1]) ** 2, 1)
        phase_diff = np.abs(np.angle(pol[0]) - np.angle(pol[1]))
        if np.isclose(phase_diff, np.pi / 2):
            polarization = "circular"
        else:
            polarization = "linear"
        super().__init__(self.lasy_profile.lambda0, polarization)
        self._t_start = t_start
        self._smooth_edges = smooth_edges
        self._apply_gaussian_filter = apply_gaussian_filter
        self._gaussian_filter_sigma = gaussian_filter_sigma

    def _envelope_function(self, xi, r, z_pos):
        # Change from Wake-T to Lasy coordinates:
        # The right edge of the Wake-T grid corresponds
        # to the left edge of the Lasy grid.
        xi_max = self.solver_params["zmax"]
        t_min_0 = self.lasy_profile.axes["t"][0]
        t = (xi_max - xi) / ct.c + t_min_0 - self._t_start
        t_min = np.min(t)
        t_max = np.max(t)
        r_min = np.min(r)
        r_max = np.max(r)

        # Create Lasy laser
        laser = Laser(
            dim="rt",
            lo=(r_min, t_min),
            hi=(r_max, t_max),
            npoints=(xi.shape[1], xi.shape[0]),
            profile=self.lasy_profile,
            n_azimuthal_modes=1,
        )

        # Get vector potential from the electric field envelope.
        a_env = field_to_vector_potential(laser.grid, laser.profile.omega0)

        # Get 2D slice and change to Wake-T ordering.
        a_env = a_env[0].T[::-1]

        # Apply Gaussian filter.
        if self._apply_gaussian_filter:
            a_env = gaussian_filter(a_env, self._gaussian_filter_sigma)

        # Smooth radial edges of profile.
        if self._smooth_edges:
            r_smooth = min(np.max(self.lasy_profile.axes["r"]), np.max(r))
            a_env *= np.exp(-2 * (r / (r_smooth * 0.85)) ** 8)

        return a_env



"""
This module contains the methods for depositing the weights of a 1D slice
of plasma particles in a 2D r-z grid.

"""

import math

from wake_t.utilities.numba import njit_serial


@njit_serial()
def deposit_plasma_particles(
    z_cell, r, w, z_min, r_min, nz, nr, dz, dr, deposition_array, p_shape="cubic"
):
    """
    Deposit the the weight of a 1D slice of plasma particles into a 2D
    r-z grid.

    Parameters
    ----------
    z_cell : float
        Index of the current longitudinal position of the plasma slice.
    r : array
        Arrays containing the radial coordinates of the
        particles.
    w : array
        Weight of the particles (any quantity) which will be deposited into
        the grid.
    z_min : float
        Position of the first field value along z.
    r_min : float
        Position of the first field value along r.
    nz, nr : int
        Number of grid cells (excluding guard cells) along the longitudinal
        and radial directions.
    dz, dr : float
        Grid step size along the longitudinal and radial direction.
    deposition_array : array
        The 2D array of size (nr+4, nz+4) (including two guard cells at each
        boundary) into which the weight will be deposited (will be
        modified within this function).
    p_shape : str
        Particle shape to be used. Possible values are 'linear' or 'cubic'.

    """
    if p_shape == "linear":
        return deposit_plasma_particles_linear(
            z_cell, r, w, z_min, r_min, nz, nr, dz, dr, deposition_array
        )
    elif p_shape == "cubic":
        return deposit_plasma_particles_cubic(
            z_cell, r, w, z_min, r_min, nz, nr, dz, dr, deposition_array
        )


@njit_serial
def deposit_plasma_particles_linear(
    z_cell, r, q, z_min, r_min, nz, nr, dz, dr, deposition_array
):
    """Calculate charge distribution assuming linear particle shape."""

    r_max = nr * dr

    # Loop over particles.
    for i in range(r.shape[0]):
        # Get particle components.
        r_i = r[i]
        w_i = q[i]

        # Deposit only if particle is within field boundaries.
        if r_i <= r_max:
            # Positions of the particles in cell units.
            r_cell = (r_i - r_min) / dr

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = min(int(math.ceil(r_cell)) + 1, nr + 2)
            iz_cell = int(math.ceil(z_cell)) + 1

            # u_r: particle position wrt left neighbor gridpoint in r.
            if r_cell < 0:
                # Force all charge to be deposited above axis.
                u_r = 1.0
            elif r_cell > nr - 1:
                # Force all charge to be deposited below r_max.
                u_r = 0.0
            else:
                u_r = r_cell - int(math.ceil(r_cell)) + 1

            # u_z: particle position wrt left neighbor gridpoint in z.
            if z_cell < 0:
                # Force all charge to be deposited above z_min.
                u_z = 1.0
            elif r_cell > nz - 1:
                # Force all charge to be deposited below z_max.
                u_z = 0.0
            else:
                u_z = z_cell - int(math.ceil(z_cell)) + 1

            # Precalculate quantities.
            zsl_0 = 1.0 - u_z
            zsl_1 = u_z
            rsl_0 = 1.0 - u_r
            rsl_1 = u_r

            # Add contribution of particle to charge distribution.
            deposition_array[iz_cell + 0, ir_cell + 0] += zsl_0 * rsl_0 * w_i
            deposition_array[iz_cell + 0, ir_cell + 1] += zsl_0 * rsl_1 * w_i
            deposition_array[iz_cell + 1, ir_cell + 0] += zsl_1 * rsl_0 * w_i
            deposition_array[iz_cell + 1, ir_cell + 1] += zsl_1 * rsl_1 * w_i


@njit_serial
def deposit_plasma_particles_cubic(
    z_cell, r, q, z_min, r_min, nz, nr, dz, dr, deposition_array
):
    """Calculate charge distribution assuming cubic particle shape."""

    r_max = nr * dr

    # Loop over particles.
    for i in range(r.shape[0]):
        # Get particle components.
        r_i = r[i]
        w_i = q[i]

        # Deposit only if particle is within field boundaries.
        if r_i <= r_max:
            # Positions of the particle in cell units.
            r_cell = (r_i - r_min) / dr

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = min(int(math.ceil(r_cell)), nr + 2)
            iz_cell = int(math.ceil(z_cell))

            # Particle position wrt left neighbor gridpoint.
            u_z = z_cell - int(math.ceil(z_cell)) + 1
            u_r = r_cell - int(math.ceil(r_cell)) + 1

            # Precalculate quantities for shape coefficients.
            inv_6 = 1.0 / 6.0
            v_z = 1.0 - u_z
            v_r = 1.0 - u_r

            # Cubic particle shape coefficients in z and r.
            zsc_0 = inv_6 * v_z**3
            zsc_1 = inv_6 * (3.0 * u_z**3 - 6.0 * u_z**2 + 4.0)
            zsc_2 = inv_6 * (3.0 * v_z**3 - 6.0 * v_z**2 + 4.0)
            zsc_3 = inv_6 * u_z**3
            rsc_0 = inv_6 * v_r**3
            rsc_1 = inv_6 * (3.0 * u_r**3 - 6.0 * u_r**2 + 4.0)
            rsc_2 = inv_6 * (3.0 * v_r**3 - 6.0 * v_r**2 + 4.0)
            rsc_3 = inv_6 * u_r**3

            # Force all charge to be deposited within boundaries.
            # Below axis:
            if r_cell <= 0.0:
                rsc_3 += rsc_0
                rsc_2 += rsc_1
                rsc_0 = 0.0
                rsc_1 = 0.0
            elif r_cell <= 1.0:
                rsc_1 += rsc_0
                rsc_0 = 0.0
            # Above r_max:
            elif r_cell > nr - 1:
                rsc_0 += rsc_3
                rsc_1 += rsc_2
                rsc_2 = 0.0
                rsc_3 = 0.0
            elif r_cell > nr - 2:
                rsc_2 += rsc_3
                rsc_3 = 0.0
            # Below z_min:
            if z_cell <= 0.0:
                zsc_3 += zsc_0
                zsc_2 += zsc_1
                zsc_0 = 0.0
                zsc_1 = 0.0
            elif z_cell <= 1.0:
                zsc_1 += zsc_0
                zsc_0 = 0.0
            # Above z_max:
            elif z_cell > nz - 1:
                zsc_0 += zsc_3
                zsc_1 += zsc_2
                zsc_2 = 0.0
                zsc_3 = 0.0
            elif z_cell > nz - 2:
                zsc_2 += zsc_3
                zsc_3 = 0.0

            # Add contribution of particle to charge distribution.
            deposition_array[iz_cell + 0, ir_cell + 0] += zsc_0 * rsc_0 * w_i
            deposition_array[iz_cell + 0, ir_cell + 1] += zsc_0 * rsc_1 * w_i
            deposition_array[iz_cell + 0, ir_cell + 2] += zsc_0 * rsc_2 * w_i
            deposition_array[iz_cell + 0, ir_cell + 3] += zsc_0 * rsc_3 * w_i
            deposition_array[iz_cell + 1, ir_cell + 0] += zsc_1 * rsc_0 * w_i
            deposition_array[iz_cell + 1, ir_cell + 1] += zsc_1 * rsc_1 * w_i
            deposition_array[iz_cell + 1, ir_cell + 2] += zsc_1 * rsc_2 * w_i
            deposition_array[iz_cell + 1, ir_cell + 3] += zsc_1 * rsc_3 * w_i
            deposition_array[iz_cell + 2, ir_cell + 0] += zsc_2 * rsc_0 * w_i
            deposition_array[iz_cell + 2, ir_cell + 1] += zsc_2 * rsc_1 * w_i
            deposition_array[iz_cell + 2, ir_cell + 2] += zsc_2 * rsc_2 * w_i
            deposition_array[iz_cell + 2, ir_cell + 3] += zsc_2 * rsc_3 * w_i
            deposition_array[iz_cell + 3, ir_cell + 0] += zsc_3 * rsc_0 * w_i
            deposition_array[iz_cell + 3, ir_cell + 1] += zsc_3 * rsc_1 * w_i
            deposition_array[iz_cell + 3, ir_cell + 2] += zsc_3 * rsc_2 * w_i
            deposition_array[iz_cell + 3, ir_cell + 3] += zsc_3 * rsc_3 * w_i



import numpy as np
from numpy.testing import assert_almost_equal
import scipy.constants as ct
import matplotlib.pyplot as plt
from wake_t.physics_models.laser.laser_pulse import GaussianPulse
from wake_t.physics_models.laser.utils import unwrap


def test_gaussian_laser_in_vacuum(plot=False):
    """Test evolution of Gaussian laser pulse in vacuum.

    This test evolves a Gaussian laser pulse over a 1cm vacuum. It checks
    that the evolution is in agreement with the analytical expectation and that
    the final value of the sum of `a_mod` has not changed.
    """
    # Grid properties.
    xi_max = 20e-6
    xi_min = -20e-6
    r_max = 400e-6
    nxi = 201
    nr = 400
    dr = r_max / nr

    # Time steps.
    z_tot = 1e-2
    t_tot = z_tot / ct.c
    nt = 1000
    dt = t_tot / nt

    # Plasma susceptibility (zero in vaccum) and normalization density.
    chi = np.zeros((nxi, nr))
    n_p = 1e23

    # Laser parameters in SI units
    tau = 25e-15  # s
    w_0 = 50e-6  # m
    l_0 = 0.8e-6  # m
    z_c = 0.0  # m
    a_0 = 3
    z_foc = z_tot / 2

    # Create and initialize laser pulse.
    laser = GaussianPulse(
        z_c, l_0=l_0, w_0=w_0, a_0=a_0, tau=tau, z_foc=z_foc, polarization="circular"
    )
    laser.set_envelope_solver_params(xi_min, xi_max, r_max, nxi, nr, dt)
    laser.initialize_envelope()

    # Preallocate arrays
    laser_w = np.zeros(nt + 1)
    laser_a = np.zeros(nt + 1)

    # Store initial laser properties.
    a_env = laser.get_envelope()
    laser_w[0] = calculate_spot_size(a_env, dr)
    laser_a[0] = calculate_a0(a_env)

    # Evolve laser.
    for n in range(nt):
        laser.evolve(chi, n_p)
        a_env = laser.get_envelope()
        laser_w[n + 1] = calculate_spot_size(a_env, dr)
        laser_a[n + 1] = calculate_a0(a_env)

    # Calculate analytical evolution.
    z = np.linspace(0, z_tot, nt + 1)
    rayleigh_length = ct.pi * w_0**2 / l_0
    laser_w_an = w_0 * np.sqrt(1 + ((z - z_foc) / rayleigh_length) ** 2)
    laser_a_an = a_0 / np.sqrt(1 + ((z - z_foc) / rayleigh_length) ** 2)

    # Check that evolution is as expected.
    diff_w = np.max(np.abs(laser_w - laser_w_an) / laser_w_an)
    diff_a = np.max(np.abs(laser_a - laser_a_an) / laser_a_an)

    assert diff_a < 1e-3
    assert diff_w < 1e-3

    # Check that solution hasn't changed.
    a_mod = np.abs(a_env)
    assert_almost_equal(np.sum(a_mod), 7500.380522059235, decimal=10)

    # Make plots.
    if plot:
        plt.figure()
        plt.plot(z, laser_w)
        plt.plot(z, laser_w_an)

        plt.figure()
        plt.plot(z, laser_a)
        plt.plot(z, laser_a_an)

        plt.show()


def test_gaussian_laser_in_vacuum_with_subgrid(plot=False):
    """Test evolution of Gaussian laser pulse in vacuum using subgridding.

    This test evolves a Gaussian laser pulse over a 1cm vacuum. It checks
    that the evolution is in agreement with the analytical expectation and that
    the final value of the sum of `a_mod` has not changed. It also checks that
    the dimensions of the envelope arrays are as expected when using a subgrid.
    """
    # Grid properties.
    xi_max = 20e-6
    xi_min = -20e-6
    r_max = 400e-6
    nxi = 201
    nr = 400
    dr = r_max / nr

    # Subgrid properties
    subgrid_nxi = 400
    subgrid_nr = 900

    # Time steps.
    z_tot = 1e-2
    t_tot = z_tot / ct.c
    nt = 1000
    dt = t_tot / nt

    # Plasma susceptibility (zero in vaccum) and normalization density.
    chi = np.zeros((nxi, nr))
    n_p = 1e23

    # Laser parameters in SI units
    tau = 25e-15  # s
    w_0 = 50e-6  # m
    l_0 = 0.8e-6  # m
    z_c = 0.0  # m
    a_0 = 3
    z_foc = z_tot / 2

    # Create and initialize laser pulse.
    laser = GaussianPulse(
        z_c, l_0=l_0, w_0=w_0, a_0=a_0, tau=tau, z_foc=z_foc, polarization="circular"
    )
    laser.set_envelope_solver_params(
        xi_min,
        xi_max,
        r_max,
        nxi,
        nr,
        dt,
        subgrid_nz=subgrid_nxi,
        subgrid_nr=subgrid_nr,
    )
    laser.initialize_envelope()

    # Preallocate arrays
    laser_w = np.zeros(nt + 1)
    laser_a = np.zeros(nt + 1)

    # Store initial laser properties.
    a_env = laser.get_envelope()
    laser_w[0] = calculate_spot_size(a_env, dr)
    laser_a[0] = calculate_a0(a_env)

    # Evolve laser.
    for n in range(nt):
        laser.evolve(chi, n_p)
        a_env = laser.get_envelope()
        laser_w[n + 1] = calculate_spot_size(a_env, dr)
        laser_a[n + 1] = calculate_a0(a_env)

    # Check that the dimensions of the arrays are as expected.
    assert a_env.shape == (nxi, nr)
    assert laser._a_env.shape == (subgrid_nxi + 2, subgrid_nr)

    # Calculate analytical evolution.
    z = np.linspace(0, z_tot, nt + 1)
    rayleigh_length = ct.pi * w_0**2 / l_0
    laser_w_an = w_0 * np.sqrt(1 + ((z - z_foc) / rayleigh_length) ** 2)
    laser_a_an = a_0 / np.sqrt(1 + ((z - z_foc) / rayleigh_length) ** 2)

    # Check that evolution is as expected.
    diff_w = np.max(np.abs(laser_w - laser_w_an) / laser_w_an)
    diff_a = np.max(np.abs(laser_a - laser_a_an) / laser_a_an)

    assert diff_a < 1e-3
    assert diff_w < 1e-3

    # Check that solution hasn't changed.
    a_mod = np.abs(a_env)
    assert_almost_equal(np.sum(a_mod), 7500.120208299948, decimal=10)

    # Make plots.
    if plot:
        plt.figure()
        plt.plot(z, laser_w)
        plt.plot(z, laser_w_an)

        plt.figure()
        plt.plot(z, laser_a)
        plt.plot(z, laser_a_an)

        plt.show()


def test_unwrap():
    """Test that the custom function for phase unwrapping implemented for
    numba agrees with the numpy version.
    """
    phase = np.linspace(0, np.pi, num=5)
    phase[3:] += np.pi
    np.testing.assert_allclose(np.unwrap(phase), unwrap(phase))


def calculate_spot_size(a_env, dr):
    # Project envelope to r
    a_proj = np.sum(np.abs(a_env), axis=0)

    # Maximum is on axis
    a_max = a_proj[0]

    # Get first index of value below a_max / e
    i_first = np.where(a_proj <= a_max / np.e)[0][0]

    # Do linear interpolation to get more accurate value of w.
    # We build a line y = a + b*x, where:
    #     b = (y_2 - y_1) / (x_2 - x_1)
    #     a = y_1 - b*x_1
    #
    #     y_1 is the value of a_proj at i_first - 1
    #     y_2 is the value of a_proj at i_first
    #     x_1 and x_2 are the radial positions of y_1 and y_2
    #
    # We can then determine the spot size by interpolating between y_1 and y_2,
    # that is, do x = (y - a) / b, where y = a_max/e
    y_1 = a_proj[i_first - 1]
    y_2 = a_proj[i_first]
    x_1 = (i_first - 1) * dr + dr / 2
    x_2 = i_first * dr + dr / 2
    b = (y_2 - y_1) / (x_2 - x_1)
    a = y_1 - b * x_1
    w = (a_max / np.e - a) / b
    return w


def calculate_a0(a_env):
    a_mod = np.abs(a_env)
    a0 = np.max(a_mod)
    return a0


if __name__ == "__main__":
    test_gaussian_laser_in_vacuum(plot=True)
    test_gaussian_laser_in_vacuum_with_subgrid(plot=True)
    test_unwrap()



from typing import Optional, Union, Callable, List, Literal

import scipy.constants as ct

from wake_t.diagnostics import OpenPMDDiagnostics
from wake_t.tracking.tracker import Tracker
from wake_t.fields.base import Field
from wake_t.particles.particle_bunch import ParticleBunch


class FieldElement:
    """
    Generic class for any beamline element based on field tracking.

    Parameters
    ----------
    length : float or str
        Length of the plasma stage in m.
    dt_bunch : float, str or list of float and str
        The time step for evolving the particle bunches. An adaptive time
        step can be used if this parameter is set to ``'auto'`` and a
        ``auto_dt_bunch`` function is provided. A list of values can
        also be provided. In this case, the list should have the same order as
        the list of bunches given to the ``track`` method.
    bunch_pusher : str
        The pusher used to evolve the particle bunches in time within
        the specified fields. Possible values are ``'rk4'`` (Runge-Kutta
        method of 4th order) or ``'boris'`` (Boris method).
    n_out : int
        Number of times along the stage in which the particle distribution
        should be returned (A list with all output bunches is returned
        after tracking).
    name : str
        Name of the plasma stage. This is only used for displaying the
        progress bar during tracking. By default, ``'field element'``.
    fields : list
        List of Fields that will be applied to the particle bunches.
    auto_dt_bunch : callable, optional
        Function used to determine the adaptive time step for bunches in
        which the time step is set to ``'auto'``. The function should take
        solely a ``ParticleBunch`` as argument.
    push_bunches_before_diags : bool, optional
        Whether to push the bunches before saving them to the diagnostics.
        Since the time step of the diagnostics can be different from that
        of the bunches, it could happen that the bunches appear in the
        diagnostics as they were at the last push, but not at the actual
        time of the diagnostics. Setting this parameter to ``True``
        (default) ensures that an additional push is given to all bunches
        to evolve them to the diagnostics time before saving.
        This additional push will always have a time step smaller than
        the the time step of the bunch, so it has no detrimental impact
        on the accuracy of the simulation. However, it could make
        convergence studies more difficult to interpret,
        since the number of pushes will depend on `n_diags`. Therefore,
        it is exposed as an option so that it can be disabled if needed.

    """

    def __init__(
        self,
        length: float,
        dt_bunch: Union[float, str, List[Union[float, str]]],
        bunch_pusher: Optional[Literal["boris", "rk4"]] = "boris",
        n_out: Optional[int] = 1,
        name: Optional[str] = "field element",
        fields: Optional[List[Field]] = [],
        auto_dt_bunch: Optional[Callable[[ParticleBunch], float]] = None,
        push_bunches_before_diags: Optional[bool] = True,
    ) -> None:
        self.length = length
        self.bunch_pusher = bunch_pusher
        self.dt_bunch = dt_bunch
        self.n_out = n_out
        self.name = name
        self.fields = fields
        self.auto_dt_bunch = auto_dt_bunch
        self.push_bunches_before_diags = push_bunches_before_diags

    def track(
        self,
        bunches: Optional[Union[ParticleBunch, List[ParticleBunch]]] = [],
        opmd_diag: Optional[Union[bool, OpenPMDDiagnostics]] = False,
        diag_dir: Optional[str] = None,
        show_progress_bar: Optional[bool] = True,
    ) -> Union[List[ParticleBunch], List[List[ParticleBunch]]]:
        """
        Track bunch through element.

        Parameters
        ----------
        bunches : ParticleBunch or list of ParticleBunch
            Particle bunches to be tracked.
        opmd_diag : bool or OpenPMDDiagnostics
            Determines whether to write simulation diagnostics to disk (i.e.
            particle distributions and fields). The output is written to
            HDF5 files following the openPMD standard. The number of outputs
            the `n_out` value. It is also possible to provide an already
            existing OpenPMDDiagnostics instance instead of a boolean value.
        diag_dir : str
            Directory into which the openPMD output will be written. By default
            this is a 'diags' folder in the current directory. Only needed if
            `opmd_diag=True`.
        show_progress_bar : bool, optional
            Whether to show a progress bar of the tracking. By default
            ``True``.

        Returns
        -------
        A list of size 'n_out' containing the bunch distribution at each step.

        """
        # Make sure `bunches` and `dt_bunch` are lists.
        if not isinstance(bunches, list):
            bunches = [bunches]
        if not isinstance(self.dt_bunch, list):
            dt_bunch = [self.dt_bunch] * len(bunches)
        else:
            n_dt, n_bunches = len(self.dt_bunch), len(bunches)
            if n_dt != n_bunches:
                raise ValueError(
                    f"The number of time steps in `dt_bunch` ({n_dt}) "
                    f"does not match the number of bunches ({n_bunches})."
                )
            dt_bunch = self.dt_bunch

        # Create diagnostics instance.
        if type(opmd_diag) is not OpenPMDDiagnostics and opmd_diag:
            opmd_diag = OpenPMDDiagnostics(write_dir=diag_dir)
        elif not opmd_diag:
            opmd_diag = None

        # Create tracker.
        tracker = Tracker(
            t_final=self.length / ct.c,
            bunches=bunches,
            dt_bunches=dt_bunch,
            fields=self.fields,
            n_diags=self.n_out,
            opmd_diags=opmd_diag,
            bunch_pusher=self.bunch_pusher,
            auto_dt_bunch_f=self.auto_dt_bunch,
            push_bunches_before_diags=self.push_bunches_before_diags,
            show_progress_bar=show_progress_bar,
            section_name=self.name,
        )

        # Do tracking.
        bunch_list = tracker.do_tracking()

        # If only tracking one bunch, do not return list of lists.
        if len(bunch_list) == 1:
            bunch_list = bunch_list[0]

        return bunch_list



import os
from copy import deepcopy
from numpy.testing import assert_array_equal
from wake_t import Beamline, Drift, PlasmaStage
from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_size


tests_output_folder = "./tests_output"


def test_single_element():
    """
    This test checks that tracking a beamline made up of a single element
    produces the same result as tracking the element in itself.

    """
    output_folder = os.path.join(tests_output_folder, "beamline_test_single")

    # Define bunch.
    bunch = get_gaussian_bunch_from_size(
        1e-6, 1e-6, 5e-6, 5e-6, 1000, 1, 10, 0, 100, 1e4
    )

    # Define plasma stage.
    plasma = PlasmaStage(
        1e-2,
        1e23,
        wakefield_model="quasistatic_2d",
        n_out=5,
        xi_min=-20e-6,
        xi_max=20e-6,
        r_max=50e-6,
        n_xi=80,
        n_r=100,
        dz_fields=0.5e-3,
    )

    # Define single-element beamline
    bl = Beamline([deepcopy(plasma)])

    # Track plasma.
    bunch_1 = bunch.copy()
    out_dir_1 = os.path.join(output_folder, "plasma_diags")
    plasma.track(bunch_1, opmd_diag=True, diag_dir=out_dir_1)

    # Track beamline.
    bunch_2 = bunch.copy()
    out_dir_2 = os.path.join(output_folder, "bl_diags")
    bl.track(bunch_2, opmd_diag=True, diag_dir=out_dir_2)

    # Check that final beams are identical.
    assert_array_equal(bunch_2.x, bunch_1.x)
    assert_array_equal(bunch_2.y, bunch_1.y)
    assert_array_equal(bunch_2.xi, bunch_1.xi)
    assert_array_equal(bunch_2.px, bunch_1.px)
    assert_array_equal(bunch_2.py, bunch_1.py)
    assert_array_equal(bunch_2.pz, bunch_1.pz)

    # Get list of output files.
    plasma_diag_files = os.listdir(os.path.join(out_dir_1, "hdf5"))
    bl_diag_files = os.listdir(os.path.join(out_dir_2, "hdf5"))

    # Check that both PlasmaStage and Beamline generate same output files.
    assert plasma_diag_files == bl_diag_files


def test_multiple_element():
    """
    This test checks that tracking a beamline made up of multiple elements
    produces the same result as tracking each element separately.

    """
    output_folder = os.path.join(tests_output_folder, "beamline_test_multi")

    # Define bunch.
    bunch = get_gaussian_bunch_from_size(
        1e-6, 1e-6, 5e-6, 5e-6, 1000, 1, 10, 0, 100, 1e4
    )

    # Define drifts.
    d1 = Drift(1e-2, n_out=3)
    d2 = Drift(1e-2, n_out=3)

    # Define plasma stage.
    plasma = PlasmaStage(
        1e-2,
        1e23,
        wakefield_model="quasistatic_2d",
        n_out=5,
        xi_min=-20e-6,
        xi_max=20e-6,
        r_max=50e-6,
        n_xi=80,
        n_r=100,
        dz_fields=0.5e-3,
    )

    # Define single-element beamline
    bl = Beamline([deepcopy(d1), deepcopy(plasma), deepcopy(d2)])

    # Track elements individually.
    bunch_1 = bunch.copy()
    out_dir_d1 = os.path.join(output_folder, "d1_diags")
    out_dir_plasma = os.path.join(output_folder, "plasma_diags")
    out_dir_d2 = os.path.join(output_folder, "d2_diags")
    d1.track(bunch_1, opmd_diag=True, diag_dir=out_dir_d1)
    plasma.track(bunch_1, opmd_diag=True, diag_dir=out_dir_plasma)
    d2.track(bunch_1, opmd_diag=True, diag_dir=out_dir_d2)

    # Track beamline.
    bunch_2 = bunch.copy()
    out_dir_bl = os.path.join(output_folder, "bl_diags")
    bl.track(bunch_2, opmd_diag=True, diag_dir=out_dir_bl)

    # Check that final beams are identical.
    assert_array_equal(bunch_2.x, bunch_1.x)
    assert_array_equal(bunch_2.y, bunch_1.y)
    assert_array_equal(bunch_2.xi, bunch_1.xi)
    assert_array_equal(bunch_2.px, bunch_1.px)
    assert_array_equal(bunch_2.py, bunch_1.py)
    assert_array_equal(bunch_2.pz, bunch_1.pz)

    # Get list of output files.
    d1_diag_files = os.listdir(os.path.join(out_dir_d1, "hdf5"))
    d2_diag_files = os.listdir(os.path.join(out_dir_d2, "hdf5"))
    plasma_diag_files = os.listdir(os.path.join(out_dir_plasma, "hdf5"))
    bl_diag_files = os.listdir(os.path.join(out_dir_bl, "hdf5"))

    # Check that both PlasmaStage and Beamline generate same output files.
    n_files_1 = len(d1_diag_files + plasma_diag_files + d2_diag_files)
    n_files_2 = len(bl_diag_files)
    assert n_files_1 == n_files_2


if __name__ == "__main__":
    test_single_element()
    test_multiple_element()



"""Contains the 5th order Adams–Bashforth pusher for the plasma particles."""

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial()
def evolve_plasma_ab5(
    dxi,
    r,
    pr,
    gamma,
    nabla_a2_pp,
    b_theta_0_pp,
    b_theta_pp,
    psi_pp,
    dr_psi_pp,
    dr_1,
    dr_2,
    dr_3,
    dr_4,
    dr_5,
    dpr_1,
    dpr_2,
    dpr_3,
    dpr_4,
    dpr_5,
):
    """
    Evolve the r and pr coordinates of plasma particles to the next xi step
    using an Adams–Bashforth method of 5th order.

    Parameters
    ----------
    dxi : float
        Longitudinal step.
    r, pr, gamma : ndarray
        Radial position, radial momentum, and Lorentz factor of the plasma
        particles.
    a2_pp, ..., dxi_psi_pp : ndarray
        Arrays where the value of the fields at the particle positions will
        be stored.
    dr_1, ..., dr_5 : ndarray
        Arrays containing the derivative of the radial position of the
        particles at the 5 slices previous to the next one.
    dpr_1, ..., dpr_5 : ndarray
        Arrays containing the derivative of the radial momentum of the
        particles at the 5 slices previous to the next one.
    """

    calculate_derivatives(
        pr, gamma, b_theta_0_pp, nabla_a2_pp, b_theta_pp, psi_pp, dr_psi_pp, dr_1, dpr_1
    )

    # Push radial position.
    apply_ab5(r, dxi, dr_1, dr_2, dr_3, dr_4, dr_5)

    # Push radial momentum.
    apply_ab5(pr, dxi, dpr_1, dpr_2, dpr_3, dpr_4, dpr_5)

    # Shift derivatives for next step (i.e., the derivative at step i will be
    # the derivative at step i+i in the next iteration.)
    dr_5[:] = dr_4
    dr_4[:] = dr_3
    dr_3[:] = dr_2
    dr_2[:] = dr_1
    dpr_5[:] = dpr_4
    dpr_4[:] = dpr_3
    dpr_3[:] = dpr_2
    dpr_2[:] = dpr_1

    # If a particle has crossed the axis, mirror it.
    idx_neg = np.where(r < 0.0)
    if idx_neg[0].size > 0:
        r[idx_neg] *= -1.0
        pr[idx_neg] *= -1.0
        dr_1[idx_neg] *= -1.0
        dr_2[idx_neg] *= -1.0
        dr_3[idx_neg] *= -1.0
        dr_4[idx_neg] *= -1.0
        dr_5[idx_neg] *= -1.0
        dpr_1[idx_neg] *= -1.0
        dpr_2[idx_neg] *= -1.0
        dpr_3[idx_neg] *= -1.0
        dpr_4[idx_neg] *= -1.0
        dpr_5[idx_neg] *= -1.0


@njit_serial()
def calculate_derivatives(
    pr, gamma, b_theta_0, nabla_a2, b_theta_bar, psi, dr_psi, dr, dpr
):
    """
    Calculate the derivative of the radial position and the radial momentum
    of the plasma particles at the current slice.

    Parameters
    ----------
    pr, gamma : ndarray
        Arrays containing the radial momentum and Lorentz factor of the
        plasma particles.
    b_theta_0 : ndarray
        Array containing the value of the azimuthal magnetic field from
        the beam distribution at the position of each plasma particle.
    nabla_a2 : ndarray
        Array containing the value of the gradient of the laser normalized
        vector potential at the position of each plasma particle.
    b_theta_bar : ndarray
        Array containing the value of the azimuthal magnetic field from
        the plasma at the position of each plasma particle.
    psi, dr_psi : ndarray
        Arrays containing the wakefield potential and its radial derivative
        at the position of each plasma particle.
    dr, dpr : ndarray
        Arrays where the value of the derivatives of the radial position and
        radial momentum will be stored.
    """
    # Calculate derivatives of r and pr.
    for i in range(pr.shape[0]):
        psi_i = psi[i]
        dpr[i] = (
            gamma[i] * dr_psi[i] / (1.0 + psi_i)
            - b_theta_bar[i]
            - b_theta_0[i]
            - nabla_a2[i] / (2.0 * (1.0 + psi_i))
        )
        dr[i] = pr[i] / (1.0 + psi_i)


@njit_serial()
def apply_ab5(x, dt, dx_1, dx_2, dx_3, dx_4, dx_5):
    """Apply the Adams-Bashforth method of 5th order to evolve `x`.

    Parameters
    ----------
    x : ndarray
        Array containing the variable to be advanced.
    dt : _type_
        Discretization step size.
    dx_1, dx_2, dx_3, dx_4, dx_5 : ndarray
        Arrays containing the derivatives of `x` at the five previous steps.
    """
    inv_720 = 1.0 / 720.0
    for i in range(x.shape[0]):
        x[i] += (
            dt
            * (
                1901.0 * dx_1[i]
                - 2774.0 * dx_2[i]
                + 2616.0 * dx_3[i]
                - 1274.0 * dx_4[i]
                + 251.0 * dx_5[i]
            )
            * inv_720
        )



"""This module contains the base class for plasma wakefields in r-z geometry"""

from typing import Optional, Callable, List

import numpy as np
import scipy.constants as ct

from wake_t.particles.interpolation import gather_main_fields_cyl_linear
from wake_t.utilities.other import generate_field_diag_dictionary
from .numerical_field import NumericalField
from wake_t.physics_models.laser.laser_pulse import LaserPulse


class RZWakefield(NumericalField):
    """Base class for plasma wakefields in r-z geometry.

    Parameters
    ----------
    density_function : callable
        Function of that returns the relative value of the plasma density
        at each `z` and `r` position.
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along r to calculate the wakefields.
    n_xi : int
        Number of grid elements along xi to calculate the wakefields.
    dz_fields : float, optional
        Determines how often the plasma wakefields should be updated.
        For example, if ``dz_fields=10e-6``, the plasma wakefields are
        only updated every time the simulation window advances by
        10 micron. By default ``dz_fields=xi_max-xi_min``, i.e., the
        length the simulation box.
    species_rho_diags : bool, optional
        Whether the model should save the charge density of each plasma species
        separately.
    laser : LaserPulse, optional
        Laser driver of the plasma stage.
    laser_evolution : bool, optional
        If True (default), the laser pulse is evolved
        using a laser envelope model. If False, the pulse envelope stays
        unchanged throughout the computation.
    laser_envelope_substeps : int, optional
        Number of substeps of the laser envelope solver per `dz_fields`.
        The time step of the envelope solver is therefore
        `dz_fields / c / laser_envelope_substeps`.
    laser_envelope_nxi, laser_envelope_nr : int, optional
        If given, the laser envelope will run in a grid of size
        (`laser_envelope_nxi`, `laser_envelope_nr`) instead
        of (`n_xi`, `n_r`). This allows the laser to run in a finer (or
        coarser) grid than the plasma wake. It is not necessary to specify
        both parameters. If one of them is not given, the resolution of
        the plasma grid with be used for that direction.
    laser_envelope_use_phase : bool, optional
        Determines whether to take into account the terms related to the
        longitudinal derivative of the complex phase in the envelope
        solver.
    field_diags : list, optional
        List of fields to save to openpmd diagnostics. By default ['rho', 'E',
        'B', 'a_mod', 'a_phase'].
    field_diags : list, optional
        List of particle quantities to save to openpmd diagnostics. By default
        [].
    model_name : str, optional
        Name of the wakefield model. This will be stored in the openPMD
        diagnostics.

    """

    def __init__(
        self,
        density_function: Callable[[float, float], float],
        r_max: float,
        xi_min: float,
        xi_max: float,
        n_r: int,
        n_xi: int,
        dz_fields=None,
        species_rho_diags: Optional[bool] = False,
        laser: Optional[LaserPulse] = None,
        laser_evolution: Optional[bool] = True,
        laser_envelope_substeps: Optional[int] = 1,
        laser_envelope_nxi: Optional[int] = None,
        laser_envelope_nr: Optional[int] = None,
        laser_envelope_use_phase: Optional[bool] = True,
        field_diags: Optional[List[str]] = ["rho", "E", "B", "a_mod", "a_phase", "a"],
        particle_diags: Optional[List[str]] = [],
        model_name: Optional[str] = "",
    ) -> None:
        dz_fields = xi_max - xi_min if dz_fields is None else dz_fields
        self.density_function = density_function
        self.laser = laser
        self.laser_evolution = laser_evolution
        self.laser_envelope_substeps = laser_envelope_substeps
        self.laser_envelope_nxi = laser_envelope_nxi
        self.laser_envelope_nr = laser_envelope_nr
        self.laser_envelope_use_phase = laser_envelope_use_phase
        self.species_rho_diags = species_rho_diags
        self.r_max = r_max
        self.xi_min = xi_min
        self.xi_max = xi_max
        self.n_r = n_r
        self.n_xi = n_xi
        self.dr = r_max / n_r
        self.dxi = (xi_max - xi_min) / (n_xi - 1)
        self.field_diags = field_diags
        self.particle_diags = particle_diags
        self.model_name = model_name
        # If a laser is included, make sure it is evolved for the whole
        # duration of the plasma stage. See `force_even_updates` parameter.
        super().__init__(
            dt_update=dz_fields / ct.c,
            openpmd_diag_supported=True,
            force_even_updates=laser is not None,
        )

    def _initialize_properties(self, bunches):
        # Initialize laser.
        if self.laser is not None:
            self.laser.set_envelope_solver_params(
                self.xi_min,
                self.xi_max,
                self.r_max,
                self.n_xi,
                self.n_r,
                self.dt_update,
                self.laser_envelope_substeps,
                self.laser_envelope_nxi,
                self.laser_envelope_nr,
                self.laser_envelope_use_phase,
            )
            self.laser.initialize_envelope()

        # Initialize field arrays
        self.rho = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.rho_e = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.rho_i = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.chi = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.e_z = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.e_r = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.e_t = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.b_z = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.b_r = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.b_t = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.r_fld = np.linspace(self.dr / 2, self.r_max - self.dr / 2, self.n_r)
        self.xi_fld = np.linspace(self.xi_min, self.xi_max, self.n_xi)

    def _evolve_properties(self, bunches):
        if self.laser is not None:
            # Evolve laser envelope
            if self.laser_evolution:
                self.laser.evolve(self.chi[2:-2, 2:-2], self.n_p)

    def _calculate_field(self, bunches):
        self.n_p = self.density_function(self.t * ct.c, 0.0)
        self.rho[:] = 0.0
        self.chi[:] = 0.0
        self.e_z[:] = 0.0
        self.e_r[:] = 0.0
        self.b_t[:] = 0.0
        if self.species_rho_diags:
            self.rho_e[:] = 0.0
            self.rho_i[:] = 0.0
        self._calculate_wakefield(bunches)

    def _calculate_wakefield(self, bunches):
        """To be implemented by the subclasses."""
        raise NotImplementedError

    def _gather(self, x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name):
        gather_main_fields_cyl_linear(
            self.e_r,
            self.e_z,
            self.b_t,
            self.xi_fld[0],
            self.xi_fld[-1],
            self.r_fld[0],
            self.r_fld[-1],
            self.dxi,
            self.dr,
            x,
            y,
            z,
            ex,
            ey,
            ez,
            bx,
            by,
            bz,
        )

    def _get_openpmd_diagnostics_data(self, global_time):
        # Prepare necessary data.
        fld_solver = "other"
        fld_solver_params = self.model_name
        fld_boundary = ["other"] * 4
        part_boundary = ["other"] * 4
        fld_boundary_params = ["none"] * 4
        part_boundary_params = ["none"] * 4
        current_smoothing = "none"
        charge_correction = "none"
        dr = np.abs(self.r_fld[1] - self.r_fld[0])
        dz = np.abs(self.xi_fld[1] - self.xi_fld[0])
        grid_spacing = [dr, dz]
        grid_labels = ["r", "z"]
        grid_global_offset = [0.0, global_time * ct.c + self.xi_min]
        # Cell-centered in 'r' and node centered in 'z'.
        fld_position = [0.5, 0.0]
        fld_names = []
        fld_comps = []
        fld_attrs = []
        fld_arrays = []
        rho_norm = self.n_p * (-ct.e)

        # Add requested fields to diagnostics.
        if "E" in self.field_diags:
            fld_names += ["E"]
            fld_comps += [["r", "t", "z"]]
            fld_attrs += [{}]
            fld_arrays += [
                [
                    np.ascontiguousarray(self.e_r.T[2:-2, 2:-2]),
                    np.ascontiguousarray(self.e_t.T[2:-2, 2:-2]),
                    np.ascontiguousarray(self.e_z.T[2:-2, 2:-2]),
                ]
            ]
        if "B" in self.field_diags:
            fld_names += ["B"]
            fld_comps += [["r", "t", "z"]]
            fld_attrs += [{}]
            fld_arrays += [
                [
                    np.ascontiguousarray(self.b_r.T[2:-2, 2:-2]),
                    np.ascontiguousarray(self.b_t.T[2:-2, 2:-2]),
                    np.ascontiguousarray(self.b_z.T[2:-2, 2:-2]),
                ]
            ]
        if "rho" in self.field_diags:
            fld_names += ["rho"]
            fld_comps += [None]
            fld_attrs += [{}]
            fld_arrays += [[np.ascontiguousarray(self.rho.T[2:-2, 2:-2]) * rho_norm]]
        if self.species_rho_diags:
            if "rho_e" in self.field_diags:
                fld_names += ["rho_e"]
                fld_comps += [None]
                fld_attrs += [{}]
                fld_arrays += [
                    [np.ascontiguousarray(self.rho_e.T[2:-2, 2:-2]) * rho_norm]
                ]
            if "rho_i" in self.field_diags:
                fld_names += ["rho_i"]
                fld_comps += [None]
                fld_attrs += [{}]
                fld_arrays += [
                    [np.ascontiguousarray(self.rho_i.T[2:-2, 2:-2]) * rho_norm]
                ]
        if self.laser is not None:
            if "a_mod" in self.field_diags:
                a_mod = np.abs(self.laser.get_envelope().T)
                fld_names += ["a_mod"]
                fld_comps += [None]
                fld_attrs += [{"polarization": self.laser.polarization}]
                fld_arrays += [[np.ascontiguousarray(a_mod)]]
            if "a_phase" in self.field_diags:
                a_phase = np.angle(self.laser.get_envelope().T)
                fld_names += ["a_phase"]
                fld_comps += [None]
                fld_attrs += [{}]
                fld_arrays += [[np.ascontiguousarray(a_phase)]]
            if "a" in self.field_diags:
                a = self.laser.get_envelope().T
                fld_names += ["a"]
                fld_comps += [None]
                if self.laser.polarization == "linear":
                    pol = np.array([1, 0j])
                else:
                    pol = np.array([np.sqrt(1 / 2), np.sqrt(1 / 2) * 1j])
                fld_attrs += [
                    {
                        "envelopeField": "normalized_vector_potential",
                        "angularFrequency": 2 * np.pi * ct.c / self.laser.l_0,
                        "polarization": pol,
                    }
                ]
                fld_arrays += [[np.ascontiguousarray(a)]]

        fld_comp_pos = [fld_position] * len(fld_names)

        # Generate dictionary for openPMD diagnostics.
        diag_data = generate_field_diag_dictionary(
            fld_names,
            fld_comps,
            fld_attrs,
            fld_arrays,
            fld_comp_pos,
            grid_labels,
            grid_spacing,
            grid_global_offset,
            fld_solver,
            fld_solver_params,
            fld_boundary,
            fld_boundary_params,
            part_boundary,
            part_boundary_params,
            current_smoothing,
            charge_correction,
        )

        return diag_data



from typing import Optional, Callable, List, Union, Dict
from warnings import warn

import numpy as np
from numpy.typing import ArrayLike
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from .solver import calculate_wakefields
from .b_theta_bunch import calculate_bunch_source, deposit_bunch_charge
from .adaptive_grid import AdaptiveGrid
from .utils import calculate_laser_a2
from wake_t.fields.rz_wakefield import RZWakefield
from wake_t.physics_models.laser.laser_pulse import LaserPulse
from wake_t.particles.particle_bunch import ParticleBunch
from wake_t.particles.interpolation import gather_main_fields_cyl_linear


class Quasistatic2DWakefieldIon(RZWakefield):
    """
    This class calculates the plasma wakefields using the gridless
    quasi-static model in r-z geometry originally developed by P. Baxevanis
    and G. Stupakov [1]_.

    The model implemented here includes additional features with respect to the
    original version in [1]_. Among them is the support for laser drivers,
    particle beams (instead of an analytic charge distribution), non-uniform
    and finite plasma density profiles, as well as an Adams-Bashforth pusher
    for the plasma particles (in addition the original Runke-Kutta pusher).

    As a kinetic quasi-static model in r-z geometry, it computes the plasma
    response by evolving a 1D radial plasma column from the front to the back
    of the simulation domain. A special feature of this model is that it does
    not need a grid in order to compute this evolution, and it allows the
    fields ``E`` and ``B`` to be calculated at any radial position in the
    plasma column.

    In the Wake-T implementation, a grid is used only in order to be able
    to easily interpolate the fields to the beam particles. After evolving the
    plasma column, ``E`` and ``B`` are calculated at the locations of the grid
    nodes. Similarly, the charge density ``rho`` and susceptibility ``chi``
    of the plasma are computed by depositing the charge of the plasma
    particles on the same grid. This useful for diagnostics and for evolving
    a laser pulse.

    Parameters
    ----------
    density_function : callable
        Function that returns the density value at the given position (z, r).
        This parameter is given by the ``PlasmaStage`` and does not need
        to be specified by the user.
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along `r` to calculate the wakefields.
    n_xi : int
        Number of grid elements along `xi` to calculate the wakefields.
    ppc : array_like, optional
        Number of plasma particles per radial cell. It can be a single number
        (e.g., ``ppc=2``) if the plasma should have the same number of
        particles per cell everywhere. Alternatively, a different number
        of particles per cell at different radial locations can also be
        specified. This can be useful, for example, when using adaptive grids
        with very narrow beams that might require more plasma particles close
        to the axis. To achieve this, an array-like structure should be given
        where each item contains two values: the number of particles per cell
        and the radius up to which this number should be used. For example
        to have 8 ppc up to a radius of 100µm and 2 ppc for higher radii up to
        500µm ``ppc=[[100e-6, 8], [500e-6, 2]]``. When using this step option
        for ``ppc`` the ``r_max_plasma`` argument is ignored. By default
        ``ppc=2``.
    dz_fields : float, optional
        Determines how often the plasma wakefields should be updated.
        For example, if ``dz_fields=10e-6``, the plasma wakefields are
        only updated every time the simulation window advances by
        10 micron. By default ``dz_fields=xi_max-xi_min``, i.e., the
        length the simulation box.
    r_max_plasma : float, optional
        Maximum radial extension of the plasma column. If ``None``, the
        plasma extends up to the ``r_max`` boundary of the simulation box.
    p_shape : str, optional
        Particle shape to be used for the beam charge deposition. Possible
        values are ``'linear'`` or ``'cubic'`` (default).
    max_gamma : float, optional
        Plasma particles whose ``gamma`` exceeds ``max_gamma`` are
        considered to violate the quasistatic condition and are put at
        rest (i.e., ``gamma=1.``, ``pr=pz=0.``). By default
        ``max_gamma=10``.
    plasma_pusher : str, optional
        The pusher used to evolve the plasma particles. Possible values
        are ``'ab2'`` (Adams-Bashforth 2nd order).
    ion_motion : bool, optional
        Whether to allow the plasma ions to move. By default, False.
    ion_mass : float, optional
        Mass of the plasma ions. By default, the mass of a proton.
    free_electrons_per_ion : int, optional
        Number of free electrons per ion. The ion charge is adjusted
        accordingly to maintain a quasi-neutral plasma (i.e.,
        ion charge = e * free_electrons_per_ion). By default, 1.
    laser : LaserPulse, optional
        Laser driver of the plasma stage.
    laser_evolution : bool, optional
        If True (default), the laser pulse is evolved
        using a laser envelope model. If ``False``, the pulse envelope
        stays unchanged throughout the computation.
    laser_envelope_substeps : int, optional
        Number of substeps of the laser envelope solver per ``dz_fields``.
        The time step of the envelope solver is therefore
        ``dz_fields / c / laser_envelope_substeps``.
    laser_envelope_nxi, laser_envelope_nr : int, optional
        If given, the laser envelope will run in a grid of size
        (``laser_envelope_nxi``, ``laser_envelope_nr``) instead
        of (``n_xi``, ``n_r``). This allows the laser to run in a finer (or
        coarser) grid than the plasma wake. It is not necessary to specify
        both parameters. If one of them is not given, the resolution of
        the plasma grid with be used for that direction.
    laser_envelope_use_phase : bool, optional
        Determines whether to take into account the terms related to the
        longitudinal derivative of the complex phase in the envelope
        solver.
    field_diags : list, optional
        List of fields to save to openpmd diagnostics. By default ['rho', 'E',
        'B', 'a_mod', 'a_phase'].
    field_diags : list, optional
        List of particle quantities to save to openpmd diagnostics. By default
        [].
    use_adaptive_grids : bool, optional
        Whether to use adaptive grids for each particle bunch, instead of the
        general (n_xi x n_r) grid.
    adaptive_grid_nr : int or list of int, optional
        Radial resolution of the adaptive grids. If only one value is given,
        the same resolution will be used for the adaptive grids of all bunches.
        Otherwise, a list of values can be given (one per bunch and in the same
        order as the list of bunches given to the `track` method). If the
        value is `None`, no adaptive grid will be used for the corresponding
        bunch, which will instead use the base grid.
    adaptive_grid_r_max : float or list of float, optional
        Specify a fixed radial extent for the adaptive grids. If not given,
        the radial extent of the grids is continuously adapted with the
        transverse size of the bunches. If only one value is given,
        the same extent will be used for all adaptive grids.
        Otherwise, a list of values can be given (one per bunch and in the same
        order as the list of bunches given to the `track` method). The
        individual values can be `float` or `None` (in which case, no fixed
        radial extent is used for the corresponding grid).
    adaptive_grid_r_lim : float or list of float, optional
        Specify a limit to the radial extent of the adaptive grids. If not
        given, the radial extent of the grids is continuously adapted to fit
        the whole transverse size of the bunches. If only one value is given,
        the same limit will be used for all adaptive grids.
        Otherwise, a list of values can be given (one per bunch and in the same
        order as the list of bunches given to the `track` method). The
        individual values can be `float` or `None` (in which case, no radial
        limit is used for the corresponding grid). Bunch particles that escape
        the grid transversely with deposit to and gather from the base grid
        (if they haven't escaped from it too).
    adaptive_grid_diags : list, optional
        List of fields from the adaptive grids to save to openpmd diagnostics.
        By default ['E', 'B'].

    References
    ----------
    .. [1] P. Baxevanis and G. Stupakov, "Novel fast simulation technique
        for axisymmetric plasma wakefield acceleration configurations in
        the blowout regime," Phys. Rev. Accel. Beams 21, 071301 (2018),
        https://link.aps.org/doi/10.1103/PhysRevAccelBeams.21.071301

    """

    def __init__(
        self,
        density_function: Callable[[float, float], float],
        r_max: float,
        xi_min: float,
        xi_max: float,
        n_r: int,
        n_xi: int,
        ppc: Optional[ArrayLike] = 2,
        dz_fields: Optional[float] = None,
        r_max_plasma: Optional[float] = None,
        parabolic_coefficient: Optional[float] = None,
        p_shape: Optional[str] = "cubic",
        max_gamma: Optional[float] = 10,
        plasma_pusher: Optional[str] = "ab2",
        ion_motion: Optional[bool] = False,
        ion_mass: Optional[float] = ct.m_p,
        free_electrons_per_ion: Optional[int] = 1,
        laser: Optional[LaserPulse] = None,
        laser_evolution: Optional[bool] = True,
        laser_envelope_substeps: Optional[int] = 1,
        laser_envelope_nxi: Optional[int] = None,
        laser_envelope_nr: Optional[int] = None,
        laser_envelope_use_phase: Optional[bool] = True,
        field_diags: Optional[List[str]] = [
            "rho",
            "E",
            "B",
            "a",
        ],
        particle_diags: Optional[List[str]] = [],
        use_adaptive_grids: Optional[bool] = False,
        adaptive_grid_nr: Optional[Union[int, List[int]]] = 16,
        adaptive_grid_r_max: Optional[Union[float, List[float]]] = None,
        adaptive_grid_r_lim: Optional[Union[float, List[float]]] = None,
        adaptive_grid_diags: Optional[List[str]] = ["E", "B"],
    ) -> None:
        # Checks for backward compatibility.
        if plasma_pusher not in ["ab2"]:
            if plasma_pusher in ["rk4", "ab5"]:
                plasma_pusher = "ab2"
                warn(
                    "Plasma pusher {plasma_pusher} has been deprecated. "
                    "Using 'ab2' pusher instead.",
                    DeprecationWarning,
                )
            else:
                raise ValueError(
                    "Plasma pusher {plasma_pusher} not recognized. "
                    "Possible values are ['ab2']."
                )
        if parabolic_coefficient is not None:
            warn(
                "The 'parabolic_coefficient' parameter has been deprecated. "
                "It will still work in order to maintain backward "
                "compatibility, but the "
                "new recommended approach is to provide a density function "
                "that takes `z` and `r` as input parameters.",
                DeprecationWarning,
            )
            parabolic_coefficient = self._get_parabolic_coefficient_fn(
                parabolic_coefficient
            )
        if r_max_plasma is None:
            r_max_plasma = r_max - r_max / n_r / 2
        self.r_max_plasma = r_max_plasma
        self.ppc = np.array(ppc)
        self.p_shape = p_shape
        self.max_gamma = max_gamma
        self.plasma_pusher = plasma_pusher
        self.ion_motion = ion_motion
        self.ion_mass = ion_mass
        self.free_electrons_per_ion = free_electrons_per_ion
        self.use_adaptive_grids = use_adaptive_grids
        self.adaptive_grid_nr = adaptive_grid_nr
        self.adaptive_grid_r_max = adaptive_grid_r_max
        self.adaptive_grid_r_lim = adaptive_grid_r_lim
        self.adaptive_grid_diags = adaptive_grid_diags
        self.bunch_grids: Dict[str, AdaptiveGrid] = {}
        self._t_reset_bunch_arrays = -1.0
        if len(self.ppc.shape) in [0, 1]:
            self.ppc = np.array([[self.r_max_plasma, self.ppc.flatten()[0]]])
        self.parabolic_coefficient = parabolic_coefficient

        super().__init__(
            density_function=density_function,
            r_max=r_max,
            xi_min=xi_min,
            xi_max=xi_max,
            n_r=n_r,
            n_xi=n_xi,
            dz_fields=dz_fields,
            species_rho_diags=True,
            laser=laser,
            laser_evolution=laser_evolution,
            laser_envelope_substeps=laser_envelope_substeps,
            laser_envelope_nxi=laser_envelope_nxi,
            laser_envelope_nr=laser_envelope_nr,
            laser_envelope_use_phase=laser_envelope_use_phase,
            field_diags=field_diags,
            particle_diags=particle_diags,
            model_name="quasistatic_2d_ion",
        )

    def _initialize_properties(self, bunches):
        super()._initialize_properties(bunches)
        # Add bunch source array (needed if not using adaptive grids).
        self.b_t_bunch = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.q_bunch = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.laser_a2 = np.zeros((self.n_xi + 4, self.n_r + 4))
        self.fld_arrays = [
            self.rho,
            self.rho_e,
            self.rho_i,
            self.chi,
            self.e_r,
            self.e_z,
            self.b_t,
            self.xi_fld,
            self.r_fld,
        ]

    def _calculate_wakefield(self, bunches: List[ParticleBunch]):
        radial_density = self._get_radial_density(self.t * ct.c)

        # Get square of laser envelope
        if self.laser is not None:
            calculate_laser_a2(self.laser.get_envelope(), self.laser_a2)
            # If linearly polarized, divide by 2 so that the ponderomotive
            # force on the plasma particles is correct.
            if self.laser.polarization == "linear":
                self.laser_a2 /= 2.0
            laser_a2 = self.laser_a2
        else:
            laser_a2 = None

        # Store plasma history if required by the diagnostics.
        store_plasma_history = len(self.particle_diags) > 0

        # Initialize empty lists with correct type so that numba can use
        # them even if there are no bunch sources.
        bunch_source_arrays = []
        bunch_source_xi_indices = []
        bunch_source_metadata = []

        # Calculate bunch sources and create adaptive grids if needed.
        s_d = ge.plasma_skin_depth(self.n_p * 1e-6)
        deposit_outliers_on_base_grid = False
        if self.use_adaptive_grids:
            store_plasma_history = True
            # Get radial grid resolution.
            if isinstance(self.adaptive_grid_nr, list):
                assert len(self.adaptive_grid_nr) == len(bunches), (
                    "Several resolutions for the adaptive grids have been "
                    "given, but they do not match the number of tracked "
                    "bunches"
                )
                nr_grids = self.adaptive_grid_nr
            else:
                nr_grids = [self.adaptive_grid_nr] * len(bunches)
            # Get radial extent.
            if isinstance(self.adaptive_grid_r_max, list):
                assert len(self.adaptive_grid_r_max) == len(bunches), (
                    "Several `r_max` for the adaptive grids have been "
                    "given, but they do not match the number of tracked "
                    "bunches"
                )
                r_max_grids = self.adaptive_grid_r_max
            else:
                r_max_grids = [self.adaptive_grid_r_max] * len(bunches)
            # Get radial extent limit.
            if isinstance(self.adaptive_grid_r_lim, list):
                assert len(self.adaptive_grid_r_lim) == len(bunches), (
                    "Several `r_lim` for the adaptive grids have been "
                    "given, but they do not match the number of tracked "
                    "bunches"
                )
                r_lim_grids = self.adaptive_grid_r_lim
            else:
                r_lim_grids = [self.adaptive_grid_r_lim] * len(bunches)
            # Check that the given extents are not larger than the limits.
            for r_lim, r_max in zip(r_lim_grids, r_max_grids):
                if r_lim is not None and r_max is not None:
                    if r_max > r_lim:
                        raise ValueError("`r_max` cannot be larger than `r_lim`")
            # Create adaptive grids for each bunch.
            bunches_with_grid: List[ParticleBunch] = []
            bunches_without_grid: List[ParticleBunch] = []
            for i, bunch in enumerate(bunches):
                if nr_grids[i] is not None:
                    bunches_with_grid.append(bunch)
                    if bunch.name not in self.bunch_grids:
                        self.bunch_grids[bunch.name] = AdaptiveGrid(
                            bunch.x,
                            bunch.y,
                            bunch.xi,
                            bunch.name,
                            nr_grids[i],
                            self.n_xi,
                            self.xi_fld,
                            r_max_grids[i],
                            r_lim_grids[i],
                        )
                else:
                    bunches_without_grid.append(bunch)
            # Calculate bunch sources at each grid.
            for bunch in bunches_with_grid:
                grid = self.bunch_grids[bunch.name]
                all_deposited = grid.calculate_bunch_source(
                    bunch, self.n_p, self.p_shape
                )
                bunch_source_arrays.append(grid.b_t_bunch)
                bunch_source_xi_indices.append(grid.i_grid)
                bunch_source_metadata.append(
                    np.array([grid.r_min_cell, grid.r_max_cell_guard, grid.dr]) / s_d
                )
                if not all_deposited:
                    self._reset_bunch_arrays()
                    deposit_bunch_charge(
                        bunch.x,
                        bunch.y,
                        bunch.xi,
                        bunch.q,
                        self.n_p,
                        self.n_r,
                        self.n_xi,
                        self.r_fld,
                        self.xi_fld,
                        self.dr,
                        self.dxi,
                        self.p_shape,
                        self.q_bunch,
                        r_min_deposit=grid.r_max,
                    )
                    deposit_outliers_on_base_grid = True

        else:
            bunches_without_grid = bunches
        # If not using adaptive grids, add all sources to the same array.
        if bunches_without_grid or deposit_outliers_on_base_grid:
            self._reset_bunch_arrays()
            for bunch in bunches_without_grid:
                deposit_bunch_charge(
                    bunch.x,
                    bunch.y,
                    bunch.xi,
                    bunch.q,
                    self.n_p,
                    self.n_r,
                    self.n_xi,
                    self.r_fld,
                    self.xi_fld,
                    self.dr,
                    self.dxi,
                    self.p_shape,
                    self.q_bunch,
                )
            calculate_bunch_source(self.q_bunch, self.n_r, self.n_xi, self.b_t_bunch)
            bunch_source_arrays.append(self.b_t_bunch)
            bunch_source_xi_indices.append(np.arange(self.n_xi))
            bunch_source_metadata.append(
                np.array(
                    [
                        self.r_fld[0],
                        self.r_fld[-1] + 2 * self.dr,  # r of last guard cell.
                        self.dr,
                    ]
                )
                / s_d
            )

        # Calculate rho only if requested in the diagnostics.
        calculate_rho = any("rho" in diag for diag in self.field_diags)

        # Calculate plasma wakefields
        self.pp = calculate_wakefields(
            laser_a2,
            self.r_max,
            self.xi_min,
            self.xi_max,
            self.n_r,
            self.n_xi,
            self.ppc,
            self.n_p,
            r_max_plasma=self.r_max_plasma,
            radial_density=radial_density,
            p_shape=self.p_shape,
            max_gamma=self.max_gamma,
            plasma_pusher=self.plasma_pusher,
            ion_motion=self.ion_motion,
            ion_mass=self.ion_mass,
            free_electrons_per_ion=self.free_electrons_per_ion,
            fld_arrays=self.fld_arrays,
            bunch_source_arrays=bunch_source_arrays,
            bunch_source_xi_indices=bunch_source_xi_indices,
            bunch_source_metadata=bunch_source_metadata,
            store_plasma_history=store_plasma_history,
            calculate_rho=calculate_rho,
            particle_diags=self.particle_diags,
        )

        # Add bunch density to total density.
        if calculate_rho:
            rho_bunch = -self.q_bunch[2:-2, 2:-2] / (self.r_fld / s_d)
            self.rho[2:-2, 2:-2] += rho_bunch

        # Calculate fields on adaptive grids.
        if self.use_adaptive_grids:
            for _, grid in self.bunch_grids.items():
                grid.calculate_fields(self.n_p, self.pp)

    def _reset_bunch_arrays(self):
        """Reset to zero the bunch arrays of the base grid."""
        if self.t > self._t_reset_bunch_arrays:
            self.b_t_bunch[:] = 0.0
            self.q_bunch[:] = 0.0
            self._t_reset_bunch_arrays = self.t

    def _get_radial_density(self, z_current):
        """Get radial density profile function"""

        def radial_density(r):
            n_p = self.density_function(z_current, r)
            if self.parabolic_coefficient is not None:
                pc = self.parabolic_coefficient(z_current)
                n_p = n_p + n_p * pc * r**2
            return n_p

        return radial_density

    def _get_parabolic_coefficient_fn(self, parabolic_coefficient):
        """Get parabolic_coefficient profile function"""
        if isinstance(parabolic_coefficient, float):

            def uniform_parabolic_coefficient(z):
                return np.ones_like(z) * parabolic_coefficient

            return uniform_parabolic_coefficient
        elif callable(parabolic_coefficient):
            return parabolic_coefficient
        else:
            raise ValueError(
                "Type {} not supported for parabolic_coefficient.".format(
                    type(parabolic_coefficient)
                )
            )

    def _gather(self, x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name):
        # If using adaptive grids, gather fields from them.
        if bunch_name in self.bunch_grids:
            grid = self.bunch_grids[bunch_name]
            grid.update_if_needed(x, y, z, self.n_p, self.pp)
            all_gathered = grid.gather_fields(x, y, z, ex, ey, ez, bx, by, bz)
            # If not all particles managed to gather from adaptive grid (for
            # example, because they escaped from it), try to gather from the
            # base grid.
            if not all_gathered:
                gather_main_fields_cyl_linear(
                    self.e_r,
                    self.e_z,
                    self.b_t,
                    self.xi_fld[0],
                    self.xi_fld[-1],
                    self.r_fld[0],
                    self.r_fld[-1],
                    self.dxi,
                    self.dr,
                    x,
                    y,
                    z,
                    ex,
                    ey,
                    ez,
                    bx,
                    by,
                    bz,
                    r_min_gather=grid.r_max_cell,
                )

        # Otherwise, use base implementation.
        else:
            super()._gather(x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name)

    def _get_openpmd_diagnostics_data(self, global_time):
        diag_data = super()._get_openpmd_diagnostics_data(global_time)
        # Add fields from adaptive grids to openpmd diagnostics.
        if self.use_adaptive_grids:
            for _, grid in self.bunch_grids.items():
                grid_data = grid.get_openpmd_data(global_time, self.adaptive_grid_diags)
                diag_data["fields"] += grid_data["fields"]
                for field in grid_data["fields"]:
                    diag_data[field] = grid_data[field]
        # Add plasma particles to openpmd diagnostics.
        particle_diags = self._get_plasma_particle_diagnostics(global_time)
        diag_data = {**diag_data, **particle_diags}
        diag_data["species"] = list(particle_diags.keys())
        return diag_data

    def _get_plasma_particle_diagnostics(self, global_time):
        """Return dict with plasma particle diagnostics."""
        diag_dict = {}
        elec_name = "plasma_electrons"
        ions_name = "plasma_ions"
        if len(self.particle_diags) > 0:
            n_elec = int(self.pp["r_hist"].shape[-1] / 2)
            s_d = ge.plasma_skin_depth(self.n_p * 1e-6)
            diag_dict[elec_name] = {
                "q": -ct.e,
                "m": ct.m_e,
                "name": elec_name,
                "geometry": "rz",
            }
            diag_dict[ions_name] = {
                "q": ct.e * self.free_electrons_per_ion,
                "m": self.ion_mass,
                "name": ions_name,
                "geometry": "rz",
            }
            if "r" in self.particle_diags:
                r_e = self.pp["r_hist"][:, :n_elec] * s_d
                r_i = self.pp["r_hist"][:, n_elec:] * s_d
                diag_dict[elec_name]["r"] = r_e
                diag_dict[ions_name]["r"] = r_i
            if "z" in self.particle_diags:
                z_e = self.pp["xi_hist"][:, :n_elec] * s_d + self.xi_max
                z_i = self.pp["xi_hist"][:, n_elec:] * s_d + self.xi_max
                diag_dict[elec_name]["z"] = z_e
                diag_dict[ions_name]["z"] = z_i
                diag_dict[elec_name]["z_off"] = global_time * ct.c
                diag_dict[ions_name]["z_off"] = global_time * ct.c
            if "pr" in self.particle_diags:
                pr_e = self.pp["pr_hist"][:, :n_elec] * ct.m_e * ct.c
                pr_i = self.pp["pr_hist"][:, n_elec:] * self.ion_mass * ct.c
                diag_dict[elec_name]["pr"] = pr_e
                diag_dict[ions_name]["pr"] = pr_i
            if "pz" in self.particle_diags:
                pz_e = self.pp["pz_hist"][:, :n_elec] * ct.m_e * ct.c
                pz_i = self.pp["pz_hist"][:, n_elec:] * self.ion_mass * ct.c
                diag_dict[elec_name]["pz"] = pz_e
                diag_dict[ions_name]["pz"] = pz_i
            if "w" in self.particle_diags:
                w_e = self.pp["w_hist"][:, :n_elec] * self.n_p
                w_i = self.pp["w_hist"][:, n_elec:] * (
                    self.n_p / self.free_electrons_per_ion
                )
                diag_dict[elec_name]["w"] = w_e
                diag_dict[ions_name]["w"] = w_i
            if "r_to_x" in self.particle_diags:
                nc_e = self.pp["r_to_x_hist"][:, :n_elec]
                nc_i = self.pp["r_to_x_hist"][:, n_elec:]
                diag_dict[elec_name]["r_to_x"] = nc_e
                diag_dict[ions_name]["r_to_x"] = nc_i
            if "id" in self.particle_diags:
                id_e = self.pp["id_hist"][:, :n_elec]
                id_i = self.pp["id_hist"][:, n_elec:]
                diag_dict[elec_name]["id"] = id_e
                diag_dict[ions_name]["id"] = id_i
        return diag_dict



"""Contains the classes of all elements tracked using transfer matrices."""

from typing import Optional, Union, List
import time

import numpy as np
import scipy.constants as ct

from wake_t.particles.push.transfer_matrix import track_with_transfer_map
from wake_t.particles.particle_bunch import ParticleBunch
from wake_t.utilities.other import print_progress_bar
from wake_t.utilities.bunch_manipulation import (
    convert_to_ocelot_matrix,
    convert_from_ocelot_matrix,
    rotation_matrix_xz,
)
from wake_t.physics_models.collective_effects.csr import get_csr_calculator
from wake_t.diagnostics import OpenPMDDiagnostics


class TMElement:
    # TODO: fix backtracking issues.
    """
    Base class for all elements where the tracking is performed using
    transfer maps. It can also be used to define an arbitrary element
    that can be modeled with transfer maps of up to second order.

    Parameters
    ----------
    length : float
        Length of the drift space in meters.
    theta : float
        Bending angle that a particle with an energy given by `gamma_ref`
        would experience.
    k1 : float
        Normalized quadrupole strength in units of 1/m**2.
    k2 : float
        Normalized sextupole strength in units of 1/m**3.
    gamma_ref : float, optional
        Reference gamma (relativistic Lorentz factor) to be used for the
        tracking. If not specified, the average gamma of the particle
        bunch is used.
    csr_on : bool
        Whether to include CSR forces.
    n_out : int
        Number of times along the element in which the particle
        distribution should be returned or written to file (if openpmd
        diagnostics are activated).
    order : int
        Highest order of the transfer matrices. Possible values are `1` and
        `2`.

    """

    def __init__(
        self,
        length: Optional[float] = 0.0,
        theta: Optional[float] = 0.0,
        k1: Optional[float] = 0.0,
        k2: Optional[float] = 0.0,
        gamma_ref: Optional[float] = None,
        csr_on: Optional[bool] = False,
        n_out: Optional[int] = None,
        order: Optional[int] = 2,
    ) -> None:
        self.length = length
        self.theta = theta
        self.k1 = k1
        self.k2 = k2
        self.gamma_ref = gamma_ref
        self.csr_on = csr_on
        self.n_out = n_out
        self.order = order
        self.csr_calculator = get_csr_calculator()
        self.element_name = ""

    def track(
        self,
        bunch: ParticleBunch,
        backtrack: Optional[bool] = False,
        out_initial: Optional[bool] = False,
        opmd_diag: Optional[Union[bool, OpenPMDDiagnostics]] = False,
        diag_dir: Optional[str] = None,
        show_progress_bar: Optional[bool] = True,
    ) -> List[ParticleBunch]:
        """
        Track bunch through element.

        Parameters
        ----------
        bunch : ParticleBunch
            Particle bunch to be tracked.
        backtrack : bool
            Whether to perform the tracking backwards.
        out_initial : bool
            Whether to include the initial bunch in the output.
        opmd_diag : bool or OpenPMDDiagnostics
            Determines whether to write simulation diagnostics to disk (i.e.
            particle distributions and fields). The output is written to
            HDF5 files following the openPMD standard. The number of outputs
            the `n_out` value. It is also possible to provide an already
            existing OpenPMDDiagnostics instance instead of a boolean value.
        diag_dir : str
            Directory into which the openPMD output will be written. By default
            this is a 'diags' folder in the current directory. Only needed if
            `opmd_diag=True`.
        show_progress_bar : bool, optional
            Whether to show a progress bar of the tracking. By default
            ``True``.

        Returns
        -------
        A list of size 'n_out' containing the bunch distribution at each step.

        """
        # Convert bunch to ocelot units and reference frame
        bunch_mat, g_avg = self._get_beam_matrix_for_tracking(bunch)
        if self.gamma_ref is None:
            self.gamma_ref = g_avg

        # Add element to CSR calculator
        if self.csr_on:
            self.csr_calculator.add_lattice_element(self)
        else:
            self.csr_calculator.clear()
        # Determine track and output steps
        l_step, track_steps, output_steps = self._determine_steps()

        # Create diagnostics if needed
        if type(opmd_diag) is not OpenPMDDiagnostics and opmd_diag:
            opmd_diag = OpenPMDDiagnostics(write_dir=diag_dir)
        elif not opmd_diag:
            opmd_diag = None

        # Print output header
        if show_progress_bar:
            print("")
            print(self.element_name.capitalize())
            print("-" * len(self.element_name))
            self._print_element_properties()
            csr_string = "on" if self.csr_on else "off"
            print("CSR {}.".format(csr_string))
            print("")
            n_steps = len(track_steps)
            st_0 = "Tracking in {} step(s)... ".format(n_steps)

        # Start tracking
        start_time = time.time()
        output_bunch_list = list()
        if out_initial:
            output_bunch_list.append(bunch.copy())
            if opmd_diag is not None:
                opmd_diag.write_diagnostics(0.0, l_step / ct.c, [output_bunch_list[-1]])
        for i in track_steps:
            if show_progress_bar:
                print_progress_bar(st_0, i + 1, n_steps)
            l_curr = (i + 1) * l_step * (1 - 2 * backtrack)
            # Track with transfer matrix
            bunch_mat = track_with_transfer_map(
                bunch_mat,
                l_step,
                self.length,
                -self.theta,
                self.k1,
                self.k2,
                self.gamma_ref,
                order=self.order,
            )
            # Apply CSR
            if self.csr_on:
                self.csr_calculator.apply_csr(
                    bunch_mat, bunch.q, self.gamma_ref, self, l_curr
                )
            # Add bunch to output list
            if i in output_steps:
                new_bunch_mat = convert_from_ocelot_matrix(bunch_mat, self.gamma_ref)
                new_bunch = self._create_new_bunch(bunch, new_bunch_mat, l_curr)
                output_bunch_list.append(new_bunch)
                if opmd_diag is not None:
                    opmd_diag.write_diagnostics(
                        l_curr / ct.c, l_step / ct.c, [output_bunch_list[-1]]
                    )

        # Update bunch data
        self._update_input_bunch(bunch, bunch_mat, output_bunch_list)

        # Add element length to diagnostics position
        if opmd_diag is not None:
            opmd_diag.increase_z_pos(self.length)

        # Finalize
        if show_progress_bar:
            tracking_time = time.time() - start_time
            print("Done ({} s).".format(tracking_time))
            print("-" * 80)
        return output_bunch_list

    def _get_beam_matrix_for_tracking(self, bunch):
        bunch_mat = bunch.get_6D_matrix()
        # obtain with respect to reference displacement
        bunch_mat[0] -= bunch.x_ref
        # rotate by the reference angle so that it enters normal to the element
        if bunch.theta_ref != 0:
            rot = rotation_matrix_xz(-bunch.theta_ref)
            bunch_mat = np.dot(rot, bunch_mat)
        return convert_to_ocelot_matrix(bunch_mat, bunch.w, self.gamma_ref)

    def _determine_steps(self):
        if self.n_out is not None:
            l_step = self.length / self.n_out
            n_track = self.n_out
            frac_out = 1
        else:
            l_step = self.length
            n_track = 1
            frac_out = 0
        if self.csr_on:
            csr_track_step = self.csr_calculator.get_csr_step(self)
            n_csr = int(self.length / csr_track_step)
            if self.n_out is not None:
                frac_out = int(n_csr / self.n_out)
            l_step = csr_track_step
            n_track = n_csr
        track_steps = np.arange(0, n_track)
        if frac_out != 0:
            output_steps = track_steps[::-frac_out]
        else:
            output_steps = []
        return l_step, track_steps, output_steps

    def _update_input_bunch(self, bunch, bunch_mat, output_bunch_list):
        if len(output_bunch_list) == 0:
            new_bunch_mat = convert_from_ocelot_matrix(bunch_mat, self.gamma_ref)
            last_bunch = self._create_new_bunch(bunch, new_bunch_mat, self.length)
        else:
            last_bunch = output_bunch_list[-1].copy()
        bunch.set_phase_space(
            last_bunch.x,
            last_bunch.y,
            last_bunch.xi,
            last_bunch.px,
            last_bunch.py,
            last_bunch.pz,
        )
        bunch.prop_distance = last_bunch.prop_distance
        bunch.theta_ref = last_bunch.theta_ref
        bunch.x_ref = last_bunch.x_ref

    def _create_new_bunch(self, old_bunch, new_bunch_mat, prop_dist):
        if self.theta != 0:
            # angle rotated for prop_dist
            theta_step = self.theta * prop_dist / self.length
            # magnet bending radius
            rho = abs(self.length / self.theta)
            # new reference angle and transverse displacement
            new_theta_ref = old_bunch.theta_ref + theta_step
            sign = -theta_step / abs(theta_step)
            new_x_ref = old_bunch.x_ref + sign * rho * (
                np.cos(new_theta_ref) - np.cos(old_bunch.theta_ref)
            )
        else:
            # new reference angle and transverse displacement
            new_theta_ref = old_bunch.theta_ref
            new_x_ref = old_bunch.x_ref + prop_dist * np.sin(old_bunch.theta_ref)
        # new prop. distance
        new_prop_dist = old_bunch.prop_distance + prop_dist
        # rotate distribution if reference angle != 0
        if new_theta_ref != 0:
            rot = rotation_matrix_xz(new_theta_ref)
            new_bunch_mat = np.dot(rot, new_bunch_mat)
        new_bunch_mat[0] += new_x_ref
        # create new bunch
        new_bunch = ParticleBunch(
            old_bunch.w,
            bunch_matrix=new_bunch_mat,
            prop_distance=new_prop_dist,
            name=old_bunch.name,
            q_species=old_bunch.q_species,
            m_species=old_bunch.m_species,
        )
        new_bunch.theta_ref = new_theta_ref
        new_bunch.x_ref = new_x_ref
        return new_bunch

    def _print_element_properties(self):
        "To be implemented by each element. Prints the element properties"
        raise NotImplementedError


class Drift(TMElement):
    """
    Defines a drift space where the tracking is performed by transfer matrices
    up to second order.

    Parameters
    ----------
    length : float
        Length of the drift space in meters.
    gamma_ref : float, optional
        Reference gamma (relativistic Lorentz factor) to be used for the
        tracking. If not specified, the average gamma of the particle
        bunch is used.
    csr_on : bool
        Whether to include CSR forces.
    n_out : int
        Number of times along the element in which the particle
        distribution should be returned or written to file (if openpmd
        diagnostics are activated).
    order : int
        Highest order of the transfer matrices. Possible values are `1` and
        `2`.

    """

    def __init__(
        self,
        length: Optional[float] = 0.0,
        gamma_ref: Optional[float] = None,
        csr_on: Optional[bool] = False,
        n_out: Optional[int] = None,
        order: Optional[int] = 2,
    ) -> None:
        super().__init__(length, 0, 0, 0, gamma_ref, csr_on, n_out, order)
        self.element_name = "drift"

    def _print_element_properties(self):
        print("Length = {:1.4f} m".format(self.length))


class Dipole(TMElement):
    """
    Defines a dipole where the tracking is performed by transfer matrices
    up to second order.

    Parameters
    ----------
    length : float
        Length of the drift space in meters.
    theta : float
        Bending angle that a particle with an energy given by `gamma_ref`
        would experience.
    gamma_ref : float, optional
        Reference gamma (relativistic Lorentz factor) to be used for the
        tracking. If not specified, the average gamma of the particle
        bunch is used.
    csr_on : bool
        Whether to include CSR forces.
    n_out : int
        Number of times along the element in which the particle
        distribution should be returned or written to file (if openpmd
        diagnostics are activated).
    order : int
        Highest order of the transfer matrices. Possible values are `1` and
        `2`.

    """

    def __init__(
        self,
        length: Optional[float] = 0.0,
        theta: Optional[float] = 0.0,
        gamma_ref: Optional[float] = None,
        csr_on: Optional[bool] = False,
        n_out: Optional[int] = None,
        order: Optional[int] = 2,
    ) -> None:
        super().__init__(length, theta, 0, 0, gamma_ref, csr_on, n_out, order)
        self.element_name = "dipole"

    def _print_element_properties(self):
        ang_deg = self.theta * 180 / ct.pi
        b_field = (ct.m_e * ct.c / ct.e) * self.theta * self.gamma_ref / self.length
        print("Bending angle = {:1.4f} rad ({:1.4f} deg)".format(self.theta, ang_deg))
        print("Dipole field = {:1.4f} T".format(b_field))


class Quadrupole(TMElement):
    """
    Defines a quadrupole where the tracking is performed by transfer matrices
    up to second order.

    Parameters
    ----------
    length : float
        Length of the drift space in meters.
    k1 : float
        Normalized quadrupole strength in units of 1/m**2.
    gamma_ref : float, optional
        Reference gamma (relativistic Lorentz factor) to be used for the
        tracking. If not specified, the average gamma of the particle
        bunch is used.
    csr_on : bool
        Whether to include CSR forces.
    n_out : int
        Number of times along the element in which the particle
        distribution should be returned or written to file (if openpmd
        diagnostics are activated).
    order : int
        Highest order of the transfer matrices. Possible values are `1` and
        `2`.

    """

    def __init__(
        self,
        length: Optional[float] = 0.0,
        k1: Optional[float] = 0.0,
        gamma_ref: Optional[float] = None,
        csr_on: Optional[bool] = False,
        n_out: Optional[int] = None,
        order: Optional[int] = 2,
    ) -> None:
        super().__init__(length, 0, k1, 0, gamma_ref, csr_on, n_out, order)
        self.element_name = "quadrupole"

    def _print_element_properties(self):
        g = self.k1 * self.gamma_ref * (ct.m_e * ct.c / ct.e)
        print("Quadrupole gradient = {:1.4f} T/m".format(g))


class Sextupole(TMElement):
    """
    Defines a sextupole where the tracking is performed by transfer matrices
    up to second order.

    Parameters
    ----------
    length : float
        Length of the drift space in meters.
    k2 : float
        Normalized sextupole strength in units of 1/m^3.
    gamma_ref : float, optional
        Reference gamma (relativistic Lorentz factor) to be used for the
        tracking. If not specified, the average gamma of the particle
        bunch is used.
    csr_on : bool
        Whether to include CSR forces.
    n_out : int
        Number of times along the element in which the particle
        distribution should be returned or written to file (if openpmd
        diagnostics are activated).
    order : int
        Highest order of the transfer matrices. Possible values are `1` and
        `2`.

    """

    def __init__(
        self,
        length: Optional[float] = 0.0,
        k2: Optional[float] = 0.0,
        gamma_ref: Optional[float] = None,
        csr_on: Optional[bool] = False,
        n_out: Optional[int] = None,
        order: Optional[int] = 2,
    ) -> None:
        super().__init__(length, 0, 0, k2, gamma_ref, csr_on, n_out, order)
        self.element_name = "sextupole"

    def _print_element_properties(self):
        g = self.k2 * self.gamma_ref * (ct.m_e * ct.c / ct.e)
        print("Sextupole gradient = {:1.4f} T/m^2".format(g))



import numpy as np
import scipy.constants as ct
import matplotlib.pyplot as plt

from wake_t.particles.particle_bunch import ParticleBunch
from wake_t.fields.analytical_field import AnalyticalField
from wake_t.tracking.tracker import Tracker


def test_different_species_single_particle(make_plots=False):
    """Test that the evolution of different particle species is correct.

    This test simulates the evolution of electrons, positrons and protons in
    a constant magnetic field. It checks that the numerical results obtained
    with all particle pushers agree with the analytical solution.
    """
    # Define particles with different mass and charge.
    electron = {"m_species": ct.m_e, "q_species": -ct.e}
    positron = {"m_species": ct.m_e, "q_species": ct.e}
    proton = {"m_species": ct.m_p, "q_species": ct.e}
    species_params = [electron, positron, proton]

    # Total tracking time.
    tracking_time = 1e-12

    # Initial longitudinal momentum.
    pz = 100.0

    # Magnetic field to apply
    b_y = 10.0  # T

    def constant_by(x, y, z, t, by, constants):
        by[:] = constants[0]

    field = AnalyticalField(b_y=constant_by, constants=[b_y])

    # List of particle pushers to test.
    pushers_to_test = ["rk4", "boris"]

    # Test particle evolution with all pushers.
    for pusher in pushers_to_test:
        # Create list with all particles to track.
        particles = []
        for params in species_params:
            particle = ParticleBunch(
                w=np.array([1.0]),
                x=np.array([0.0]),
                y=np.array([0.0]),
                xi=np.array([0.0]),
                px=np.array([0.0]),
                py=np.array([0.0]),
                pz=np.array([pz]),
                **params,
            )
            particles.append(particle)

        # Create tracker.
        tracker = Tracker(
            t_final=tracking_time,
            bunches=particles,
            dt_bunches=[0.1e-15, 0.1e-15, 0.1e-15],
            fields=[field],
            n_diags=50,
            bunch_pusher=pusher,
        )

        # Track particles.
        particles_history = tracker.do_tracking()

        # Compare with analytical solution.
        for particle, params in zip(particles, species_params):
            gamma = np.sqrt(1 + pz**2)
            vz_0 = pz / gamma * ct.c
            w = -params["q_species"] * b_y / (params["m_species"] * gamma)
            x_an = vz_0 / w * (1.0 - np.cos(w * tracking_time))
            np.testing.assert_almost_equal(particle.x, x_an, decimal=14)

        # Make plot.
        if make_plots:
            fig, axes = plt.subplots(2, 1)
            for particle_history in particles_history:
                x_history = np.zeros(len(particle_history))
                y_history = np.zeros(len(particle_history))
                xi_history = np.zeros(len(particle_history))
                px_history = np.zeros(len(particle_history))
                pz_history = np.zeros(len(particle_history))
                for i, p in enumerate(particle_history):
                    x_history[i] = p.x
                    y_history[i] = p.y
                    xi_history[i] = p.xi + p.prop_distance
                    px_history[i] = p.px
                    pz_history[i] = p.pz

                axes[0].plot(xi_history, x_history)
                axes[1].plot(xi_history, px_history)
            plt.show()


if __name__ == "__main__":
    test_different_species_single_particle(make_plots=True)



"""Defines a magnetic field of a quadrupole"""

from wake_t.fields.analytical_field import AnalyticalField
from wake_t.utilities.numba import prange


def b_x(x, y, z, t, bx, constants):
    """B_x component."""
    k = -constants[0]
    for i in prange(x.shape[0]):
        bx[i] += k * y[i]


def b_y(x, y, z, t, by, constants):
    """B_y component."""
    k = -constants[0]
    for i in prange(x.shape[0]):
        by[i] += k * x[i]


class QuadrupoleField(AnalyticalField):
    """Defines a field of a magnetic quadrupole of constant focusing gradient
    `k`.

    In Cartesian coordinates, the field is given by:
    ```
        b_x = - k * y
        b_y = - k * x
    ```

    When `k > 0`, it corresponds to focussing electrons in the `x` direction
    and defocussing in `y`. When `k < 0`, the result is the opposite.

    Parameters
    ----------
    foc_gradient : float
        Uniform focusing gradient in T/m.
    """

    def __init__(self, foc_gradient):
        super().__init__(b_x=b_x, b_y=b_y, constants=[foc_gradient])



"""
This module contains the class defining a particle bunch.
"""

# TODO: clean methods to set and get bunch matrix
from __future__ import annotations
from copy import deepcopy
from typing import Optional

import numpy as np
import scipy.constants as ct
from aptools.plotting.quick_diagnostics import full_phase_space

from .push.runge_kutta_4 import apply_rk4_pusher
from .push.boris_pusher import apply_boris_pusher


class ParticleBunch:
    """Defines a particle bunch.

    Parameters
    ----------
    w : ndarray
        Weight of the macroparticles, i.e., the number of real particles
        represented by each macroparticle. In practice, :math:`w = q_m / q`,
        where :math:`q_m` and :math:`q` are, respectively the charge of the
        macroparticle and of the real particle (e.g., an electron).
    x, y, xi : ndarray
        Position of the macropparticles in the x, y, and xi directions in
        units of m.
    px, py, pz : ndarray
        Momentum of the macropparticles in the x, y, and z directions in
        non-dimensional units (beta*gamma).
    bunch_matrix : ndarray
        6 x N matrix, where N is the number of particles, containing the
        phase-space information of the bunch. If provided, the arguments x
        to pz are not considered. The matrix contains (x, px, y, py, z, pz)
        if matrix_type='standard' or (x, x', y, y', xi, dp) if
        matrix_type='alternative'.
    matrix_type : string
        Indicates the type of bunch_matrix. Possible values are 'standard'
        or 'alternative' (see above).
    gamma_ref : float
        Reference energy with respect to which the particle momentum dp is
        calculated. Only needed if bunch_matrix is used and
        matrix_type='alternative'.
    tags : ndarray
        Individual tags assigned to each particle.
    prop_distance : float
        Propagation distance of the bunch along the beamline.
    t_flight : float
        Time of flight of the bunch along the beamline.
    z_injection: float
        Particles have a ballistic motion for z<z_injection (in meters).
    name : str
        Name of the particle bunch. Used for species identification
        in openPMD diagnostics.
    q_species, m_species : float
        Charge and mass of a single particle of the species represented
        by the macroparticles. For an electron bunch (default),
        ``q_species=-e`` and ``m_species=m_e``

    """

    _n_unnamed = 0  # Number of unnamed ParticleBunch instances

    def __init__(
        self,
        w: np.ndarray,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        xi: Optional[np.ndarray] = None,
        px: Optional[np.ndarray] = None,
        py: Optional[np.ndarray] = None,
        pz: Optional[np.ndarray] = None,
        bunch_matrix: Optional[np.ndarray] = None,
        matrix_type: Optional[str] = "standard",
        gamma_ref: Optional[float] = None,
        tags: Optional[np.ndarray] = None,
        prop_distance: Optional[float] = 0,
        t_flight: Optional[float] = 0,
        z_injection: Optional[float] = None,
        name: Optional[str] = None,
        q_species: Optional[float] = -ct.e,
        m_species: Optional[float] = ct.m_e,
    ) -> None:
        if bunch_matrix is not None:
            if matrix_type == "standard":
                self.set_phase_space_from_matrix(bunch_matrix)
            elif matrix_type == "alternative":
                self.set_phase_space_from_alternative_matrix(bunch_matrix, gamma_ref)
        else:
            self.x = x
            self.y = y
            self.xi = xi
            self.px = px
            self.py = py
            self.pz = pz
        self.w = w
        # self.mu = 0
        self.tags = tags
        self.prop_distance = prop_distance
        self.t_flight = t_flight
        self.z_injection = z_injection
        self.x_ref = 0
        self.theta_ref = 0
        self.set_name(name)
        self.q_species = q_species
        self.m_species = m_species
        self.__field_arrays_allocated = False
        self.__rk4_arrays_allocated = False

    @property
    def q(self) -> np.ndarray:
        """Get an array with the charge of each macroparticle.

        This property is implemented for convenience and for backward
        compatibility.
        """
        return self.w * self.q_species

    @q.setter
    def q(self, q_new):
        """Set the total charge of each macroparticle.

        This property is implemented for convenience and for backward
        compatibility.
        """
        self.w = np.abs(q_new / self.q_species)

    def set_name(self, name):
        """Set the particle bunch name"""
        if name is None:
            name = "elec_bunch_{}".format(ParticleBunch._n_unnamed)
            ParticleBunch._n_unnamed += 1
        self.name = name

    def set_phase_space(self, x, y, xi, px, py, pz):
        """Sets the phase space coordinates"""
        self.x = x
        self.y = y
        self.xi = xi
        self.px = px
        self.py = py
        self.pz = pz

    def set_phase_space_from_matrix(self, beam_matrix):
        """
        Sets the phase space coordinates from a matrix with the values of
        (x, px, y, py, xi, pz).

        """
        self.x = beam_matrix[0]
        self.y = beam_matrix[2]
        self.xi = beam_matrix[4]
        self.px = beam_matrix[1]
        self.py = beam_matrix[3]
        self.pz = beam_matrix[5]

    def set_phase_space_from_alternative_matrix(self, beam_matrix, gamma_ref):
        """
        Sets the phase space coordinates from a matrix with the values of
        (x, x', y, y', xi, dp).

        Parameters
        ----------
        bunch_matrix : array
            6 x N matrix, where N is the number of particles, containing the
            phase-space information of the bunch as (x, x', y, y', xi, dp) in
            units of (m, rad, m, rad, m, -). dp is defined as
            dp = (g-g_ref)/g_ref, while x' = px/p_kin and y' = py/p_kin, where
            p_kin is the kinetic momentum of each particle.

        gamma_ref : float
            Reference energy with respect to which the particle momentum dp is
            calculated.

        """
        dp = beam_matrix[5]
        gamma = (dp + 1) * gamma_ref
        p_kin = np.sqrt(gamma**2 - 1)
        self.x = beam_matrix[0]
        self.px = beam_matrix[1] * p_kin
        self.y = beam_matrix[2]
        self.py = beam_matrix[3] * p_kin
        self.xi = beam_matrix[4]
        self.pz = np.sqrt(gamma**2 - self.px**2 - self.py**2 - 1)

    def set_bunch_matrix(self, beam_matrix):
        """Sets the 6D phase space and charge of the bunch"""
        self.x = beam_matrix[0]
        self.y = beam_matrix[1]
        self.xi = beam_matrix[2]
        self.px = beam_matrix[3]
        self.py = beam_matrix[4]
        self.pz = beam_matrix[5]
        self.w = beam_matrix[6] / self.q_species

    def get_bunch_matrix(self):
        """Returns a matrix with the 6D phase space and charge of the bunch"""
        return np.array(
            [
                self.x,
                self.y,
                self.xi,
                self.px,
                self.py,
                self.pz,
                self.w * self.q_species,
            ]
        )

    def get_6D_matrix(self):
        """
        Returns the 6D phase space matrix of the bunch containing
        (x, px, y, py, xi, pz)
        """
        return np.array([self.x, self.px, self.y, self.py, self.xi, self.pz])

    def get_6D_matrix_with_charge(self):
        """
        Returns the 6D phase space matrix of the bunch containing
        (x, px, y, py, xi, pz)
        """
        return np.array(
            [
                self.x,
                self.px,
                self.y,
                self.py,
                self.xi,
                self.pz,
                self.w * self.q_species,
            ]
        )

    def get_alternative_6D_matrix(self):
        """
        Returns the 6D matrix of the bunch containing
        (x, x', y, y', xi, dp)
        """
        g = np.sqrt(1 + self.px**2 + self.py**2 + self.pz**2)
        g_avg = np.average(g, weights=self.w)
        b_avg = np.sqrt(1 - g_avg ** (-2))
        dp = (g - g_avg) / (g_avg * b_avg)
        p_kin = np.sqrt(g**2 - 1)
        return np.array(
            [self.x, self.px / p_kin, self.y, self.py / p_kin, self.xi, dp]
        ), g_avg

    def increase_prop_distance(self, dist):
        """Increases the propagation distance"""
        self.prop_distance += dist

    def reposition_xi(self, xi_c):
        """Recenter bunch along xi around the specified xi_c"""
        current_xi_c = np.average(self.xi, weights=self.w)
        dxi = xi_c - current_xi_c
        self.xi += dxi

    def get_openpmd_diagnostics_data(self, global_time):
        """
        Returns a dictionary with the necessary data to write the openPMD
        diagnostics of the particle bunch.

        """
        diag_dict = {
            "x": self.x,
            "y": self.y,
            "z": self.xi,
            "px": self.px * self.m_species * ct.c,
            "py": self.py * self.m_species * ct.c,
            "pz": self.pz * self.m_species * ct.c,
            "w": self.w,
            "q": self.q_species,
            "m": self.m_species,
            "name": self.name,
            "z_off": global_time * ct.c,
            "geometry": "3d_cartesian",
        }
        if self.tags is not None:
            diag_dict["id"] = self.tags
        return diag_dict

    def show(self, **kwargs):
        """Show the phase space of the bunch in all dimensions."""
        full_phase_space(
            self.x,
            self.y,
            self.xi,
            self.px,
            self.py,
            self.pz,
            self.w * self.q_species,
            show=True,
            **kwargs,
        )

    def evolve(self, fields, t, dt, pusher="rk4"):
        """Evolve particle bunch to the next time step.

        Parameters
        ----------
        fields : list
            List of fields in which to evolve the particle bunch.
        t : float
            The current time.
        dt : float
            Time step by which to evolve the bunch.
        pusher : str, optional
            The particle pusher to use. Either 'rk4' or 'boris'. By
            default 'rk4'.
        """

        if self.z_injection is not None:
            if (np.amax(self.xi) + self.prop_distance) < self.z_injection:
                fields = []

        if pusher == "rk4":
            apply_rk4_pusher(self, fields, t, dt)
        elif pusher == "boris":
            apply_boris_pusher(self, fields, t, dt)
        else:
            raise ValueError(
                f"Bunch pusher '{pusher}' not recognized. "
                "Possible values are 'boris' and 'rk4'"
            )
        self.prop_distance += dt * ct.c

    def copy(self) -> ParticleBunch:
        """Return a copy of the bunch.

        To improve performance, this copy won't contain copies of auxiliary
        arrays, only of the particle coordinates and properties.
        """
        bunch_copy = ParticleBunch(
            w=deepcopy(self.w),
            x=deepcopy(self.x),
            y=deepcopy(self.y),
            xi=deepcopy(self.xi),
            px=deepcopy(self.px),
            py=deepcopy(self.py),
            pz=deepcopy(self.pz),
            prop_distance=deepcopy(self.prop_distance),
            name=deepcopy(self.name),
            q_species=deepcopy(self.q_species),
            m_species=deepcopy(self.m_species),
            tags=deepcopy(self.tags),
        )
        bunch_copy.x_ref = self.x_ref
        bunch_copy.theta_ref = self.theta_ref
        return bunch_copy

    def get_field_arrays(self):
        """Get the arrays where the gathered fields will be stored."""
        if not self.__field_arrays_allocated:
            self.__preallocate_field_arrays()
        return (self.__e_x, self.__e_y, self.__e_z, self.__b_x, self.__b_y, self.__b_z)

    def get_rk4_arrays(self):
        """Get the arrays needed by the RK4 pusher."""
        if not self.__rk4_arrays_allocated:
            self.__preallocate_rk4_arrays()
        return (
            self.__x_rk4,
            self.__y_rk4,
            self.__xi_rk4,
            self.__px_rk4,
            self.__py_rk4,
            self.__pz_rk4,
            self.__dx_rk4,
            self.__dy_rk4,
            self.__dxi_rk4,
            self.__dpx_rk4,
            self.__dpy_rk4,
            self.__dpz_rk4,
            self.__k_x,
            self.__k_y,
            self.__k_xi,
            self.__k_px,
            self.__k_py,
            self.__k_pz,
        )

    def __preallocate_field_arrays(self):
        """Preallocate the arrays where the gathered fields will be stored."""
        n_part = len(self.x)
        self.__e_x = np.zeros(n_part)
        self.__e_y = np.zeros(n_part)
        self.__e_z = np.zeros(n_part)
        self.__b_x = np.zeros(n_part)
        self.__b_y = np.zeros(n_part)
        self.__b_z = np.zeros(n_part)
        self.__field_arrays_allocated = True

    def __preallocate_rk4_arrays(self):
        """Preallocate the arrays needed by the RK4 pusher."""
        n_part = len(self.x)

        self.__k_x = np.zeros(n_part)
        self.__k_y = np.zeros(n_part)
        self.__k_xi = np.zeros(n_part)
        self.__k_px = np.zeros(n_part)
        self.__k_py = np.zeros(n_part)
        self.__k_pz = np.zeros(n_part)

        self.__x_rk4 = np.zeros(n_part)
        self.__y_rk4 = np.zeros(n_part)
        self.__xi_rk4 = np.zeros(n_part)
        self.__px_rk4 = np.zeros(n_part)
        self.__py_rk4 = np.zeros(n_part)
        self.__pz_rk4 = np.zeros(n_part)

        self.__dx_rk4 = np.zeros(n_part)
        self.__dy_rk4 = np.zeros(n_part)
        self.__dxi_rk4 = np.zeros(n_part)
        self.__dpx_rk4 = np.zeros(n_part)
        self.__dpy_rk4 = np.zeros(n_part)
        self.__dpz_rk4 = np.zeros(n_part)

        self.__rk4_arrays_allocated = True



"""Contains the base class for all EM fields."""

from typing import Union, Dict
import numpy as np


class Field:
    """
    Base class for all EM fields.

    It defines the interface for gathering all field components
    (i.e., Ex, Ey, Ez, Bx, By, Bz), updating the fields, and getting the
    openpmd diagnostics.
    """

    def __init__(self, openpmd_diag_supported=False):
        self.__openpmd_diag_supported = openpmd_diag_supported

    def gather(
        self,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        t: np.ndarray,
        ex: np.ndarray,
        ey: np.ndarray,
        ez: np.ndarray,
        bx: np.ndarray,
        by: np.ndarray,
        bz: np.ndarray,
        bunch_name: str,
    ) -> None:
        """Gather all field components at the specified locations.

        Parameters
        ----------
        x : ndarray
            1D array containing the x position where to gather the fields.
        y : ndarray
            1D array containing the x position where to gather the fields.
        z : ndarray
            1D array containing the x position where to gather the fields.
        t : float
            Time at which the field is being gathered.
        ex : ndarray
            1D array where the gathered Ex values will be stored.
        ey : ndarray
            1D array where the gathered Ey values will be stored
        ez : ndarray
            1D array where the gathered Ez values will be stored
        bx : ndarray
            1D array where the gathered Bx values will be stored
        by : ndarray
            1D array where the gathered By values will be stored
        bz : ndarray
            1D array where the gathered Bz values will be stored
        bunch_name : str
            Name of the bunch that is gathering the fields
        """
        self._gather(x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name)

    def get_openpmd_diagnostics_data(self, global_time: float) -> Union[Dict, None]:
        """Get the data for including the field in the openPMD diagnostics.

        Parameters
        ----------
        global_time : float
            Current global time of the simulation.

        Returns
        -------
        Dict or None
            If openPMD output is supported by the field, returns a Dict with
            all the required data.
        """
        if self.__openpmd_diag_supported:
            return self._get_openpmd_diagnostics_data(global_time)

    def _gather(self, x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name):
        """To be implemented by the subclasses."""
        raise NotImplementedError

    def _get_openpmd_diagnostics_data(self):
        """To be implemented by the subclasses."""
        raise NotImplementedError



"""
Contains the method to compute the wakefield potential and its derivatives
according to the paper by P. Baxevanis and G. Stupakov.

"""

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial(fastmath=True, error_model="numpy")
def calculate_psi_and_derivatives_at_particles(
    r_e,
    log_r_e,
    pr_e,
    w_e,
    w_center_e,
    q_e,
    r_i,
    log_r_i,
    pr_i,
    w_i,
    w_center_i,
    q_i,
    ion_motion,
    calculate_ion_sums,
    sum_1_e,
    sum_2_e,
    sum_3_e,
    sum_1_i,
    sum_2_i,
    sum_3_i,
    psi_e,
    dr_psi_e,
    dxi_psi_e,
    psi_i,
    dr_psi_i,
    dxi_psi_i,
    psi,
    dr_psi,
    dxi_psi,
):
    """Calculate wakefield potential and derivatives at the plasma particles.

    Parameters
    ----------
    r_e, log_r_e, pr_e, w_e, w_center_e : ndarray
        Radial position (and log), momentum, weight (and central weight)
        of the plasma electrons.
    q_e : float
        Charge of the plasma electron species.
    r_i, log_r_i, pr_i, w_i, w_center_i, dr_p_i : ndarray
        Radial position (and log), momentum, weight (and central weight)
        of the plasma ions.
    q_i : float
        Charge of the plasma ion species.
    ion_motion : bool
        Whether the ions can move. If `True`, the potential and its derivatives
        will also be calculated at the ions.
    calculate_ion_sums : _type_
        Whether to calculate sum_1, sum_2 and sum_3 for the ions. When ion
        motion is disabled, this only needs to be done once.
    r_neighbor_e, log_r_neighbor_e, r_neighbor_i, log_r_neighbor_i : ndarray
        The radial location (and its logarithm) of the middle points between
        each particle and it left and right neighbors. This array is already
        sorted and should not be indexed with `i_sort`.
    sum_1_e, sum_2_e, sum_3_e, sum_1_i, sum_2_i, sum_3_i : ndarray
        Arrays where the values of sum_1, sum_2 and sum_3 at each particle
        will be stored.
    psi_e, dr_psi_e, dxi_psi_e, psi_i, dr_psi_i, dxi_psi_i : ndarray
        Arrays where the value of psi and its derivatives at the plasma
        electrons and ions will be stored.
    psi, dr_psi, dxi_psi : _type_
        Arrays where the value of psi and its derivatives at all
        plasma particles is stored.
    """

    # Calculate cumulative sums 1 and 2 (Eqs. (29) and (31)).
    calculate_cumulative_sum_1(q_e, w_e, w_center_e, sum_1_e)
    calculate_cumulative_sum_2(q_e, log_r_e, w_e, w_center_e, sum_2_e)
    if ion_motion or not calculate_ion_sums:
        calculate_cumulative_sum_1(q_i, w_i, w_center_i, sum_1_i)
        calculate_cumulative_sum_2(q_i, log_r_i, w_i, w_center_i, sum_2_i)

    # Calculate the psi and dr_psi background at the neighboring points.
    # For the electrons, compute the psi and dr_psi due to the ions at
    # r_neighbor_e. For the ions, compute the psi and dr_psi due to the
    # electrons at r_neighbor_i.
    calculate_psi_and_dr_psi_at_particle_centers(
        r_e, log_r_e, sum_1_e, sum_2_e, psi_e, dr_psi_e
    )
    calculate_psi_and_dr_psi_with_interpolation(
        r_e, r_i, log_r_i, sum_1_i, sum_2_i, psi_e, dr_psi_e, add=True
    )
    if ion_motion:
        calculate_psi_and_dr_psi_at_particle_centers(
            r_i, log_r_i, sum_1_i, sum_2_i, psi_i, dr_psi_i
        )
        calculate_psi_and_dr_psi_with_interpolation(
            r_i, r_e, log_r_e, sum_1_e, sum_2_e, psi_i, dr_psi_i, add=True
        )

    # Check that the values of psi are within a reasonable range (prevents
    # issues at the peak of a blowout wake, for example).
    check_psi(psi)
    check_psi_derivative(dr_psi)

    # Calculate cumulative sum 3 (Eq. (32)).
    calculate_cumulative_sum_3(q_e, r_e, pr_e, w_e, w_center_e, psi_e, sum_3_e)
    if ion_motion or not calculate_ion_sums:
        calculate_cumulative_sum_3(q_i, r_i, pr_i, w_i, w_center_i, psi_i, sum_3_i)

    # Calculate the dxi_psi background at the neighboring points.
    # For the electrons, compute the psi and dr_psi due to the ions at
    # r_neighbor_e. For the ions, compute the psi and dr_psi due to the
    # electrons at r_neighbor_i.
    calculate_dxi_psi_at_particle_centers(r_e, sum_3_e, dxi_psi_e)
    if ion_motion:
        calculate_dxi_psi_with_interpolation(r_e, r_i, sum_3_i, dxi_psi_e, add=True)
        calculate_dxi_psi_at_particle_centers(r_i, sum_3_i, dxi_psi_i)
        calculate_dxi_psi_with_interpolation(r_i, r_e, sum_3_e, dxi_psi_i, add=True)

    # Check that the values of dxi_psi are within a reasonable range (prevents
    # issues at the peak of a blowout wake, for example).
    check_psi_derivative(dxi_psi)


@njit_serial(fastmath=True)
def calculate_cumulative_sum_1(q, w, w_center, sum_1_arr):
    """Calculate the cumulative sum in Eq. (29)."""
    sum_1 = 0.0
    for i in range(w.shape[0]):
        w_i = w[i]
        w_center_i = w_center[i]
        # Integrate up to particle centers.
        sum_1_arr[i] = sum_1 + q * w_center_i
        # And add all charge for next iteration.
        sum_1 += q * w_i
    # Total sum after last particle.
    sum_1_arr[-1] = sum_1


@njit_serial(fastmath=True)
def calculate_cumulative_sum_2(q, log_r, w, w_center, sum_2_arr):
    """Calculate the cumulative sum in Eq. (31)."""
    sum_2 = 0.0
    for i in range(log_r.shape[0]):
        log_r_i = log_r[i]
        w_i = w[i]
        w_center_i = w_center[i]
        # Integrate up to particle centers.
        sum_2_arr[i] = sum_2 + q * w_center_i * log_r_i
        # And add all charge for next iteration.
        sum_2 += q * w_i * log_r_i
    # Total sum after last particle.
    sum_2_arr[-1] = sum_2


@njit_serial(fastmath=True, error_model="numpy")
def calculate_cumulative_sum_3(q, r, pr, w, w_center, psi, sum_3_arr):
    """Calculate the cumulative sum in Eq. (32)."""
    sum_3 = 0.0
    for i in range(r.shape[0]):
        r_i = r[i]
        pr_i = pr[i]
        w_i = w[i]
        w_center_i = w_center[i]
        psi_i = psi[i]
        # Integrate up to particle centers.
        sum_3_arr[i] = sum_3 + (q * w_center_i * pr_i) / (r_i * (1 + psi_i))
        # And add all charge for next iteration.
        sum_3 += (q * w_i * pr_i) / (r_i * (1 + psi_i))
    # Total sum after last particle.
    sum_3_arr[-1] = sum_3


@njit_serial(fastmath=True, error_model="numpy")
def calculate_psi_with_interpolation(
    r_eval, r, log_r, sum_1_arr, sum_2_arr, psi, add=False
):
    """Calculate psi at the radial positions given in `r_eval`."""
    # Get number of plasma particles.
    n_part = r.shape[0]

    # Get number of points to evaluate.
    n_points = r_eval.shape[0]

    # Calculate psi after the last plasma plasma particle
    # This is used to ensure the boundary condition psi=0, which also
    # assumes that the total electron and ion charge are the same.
    sum_2_max = sum_2_arr[-1]

    # Calculate fields at r_eval.
    i_last = 0
    r_left = 0.0
    sum_1_left = 0.0
    sum_2_left = 0.0
    psi_left = 0.0
    for j in range(n_points):
        r_j = r_eval[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        while i_last < n_part:
            r_right = r[i_last]
            if r_right >= r_j:
                break
            r_left = r_right
            i_last += 1
        if i_last < n_part:
            if i_last > 0:
                log_r_left = log_r[i_last - 1]
                sum_1_left = sum_1_arr[i_last - 1]
                sum_2_left = sum_2_arr[i_last - 1]
                psi_left = sum_1_left * log_r_left - sum_2_left
            log_r_right = log_r[i_last]
            sum_1_right = sum_1_arr[i_last]
            sum_2_right = sum_2_arr[i_last]
            psi_right = sum_1_right * log_r_right - sum_2_right

            # Interpolate sums.
            inv_dr = 1.0 / (r_right - r_left)
            slope_2 = (psi_right - psi_left) * inv_dr
            psi_j = psi_left + slope_2 * (r_j - r_left) + sum_2_max
        else:
            sum_1_left = sum_1_arr[-1]
            sum_2_left = sum_2_arr[-1]
            psi_j = sum_1_left * np.log(r_j) - sum_2_left + sum_2_max

        # Calculate fields at r_j.
        if add:
            psi[j] += psi_j
        else:
            psi[j] = psi_j


@njit_serial(fastmath=True, error_model="numpy")
def calculate_psi_and_dr_psi_with_interpolation(
    r_eval, r, log_r, sum_1_arr, sum_2_arr, psi, dr_psi, add=False
):
    """Calculate psi and dr_psi at the radial positions given in `r_eval`."""
    # Get number of plasma particles.
    n_part = r.shape[0]

    # Get number of points to evaluate.
    n_points = r_eval.shape[0]

    # Calculate psi after the last plasma plasma particle
    # This is used to ensure the boundary condition psi=0, which also
    # assumes that the total electron and ion charge are the same.
    sum_2_max = sum_2_arr[-1]

    # Calculate fields at r_eval.
    i_last = 0
    r_left = 0.0
    sum_1_left = 0.0
    sum_2_left = 0.0
    psi_left = 0.0
    dr_psi_left = 0.0
    for j in range(n_points):
        r_j = r_eval[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        while i_last < n_part:
            r_right = r[i_last]
            if r_right >= r_j:
                break
            r_left = r_right
            i_last += 1
        if i_last < n_part:
            if i_last > 0:
                log_r_left = log_r[i_last - 1]
                sum_1_left = sum_1_arr[i_last - 1]
                sum_2_left = sum_2_arr[i_last - 1]
                dr_psi_left = sum_1_left / r_left
                psi_left = sum_1_left * log_r_left - sum_2_left
            log_r_right = log_r[i_last]
            sum_1_right = sum_1_arr[i_last]
            sum_2_right = sum_2_arr[i_last]
            dr_psi_right = sum_1_right / r_right
            psi_right = sum_1_right * log_r_right - sum_2_right

            # Interpolate sums.
            inv_dr = 1.0 / (r_right - r_left)
            slope_1 = (dr_psi_right - dr_psi_left) * inv_dr
            slope_2 = (psi_right - psi_left) * inv_dr
            dr_psi_j = dr_psi_left + slope_1 * (r_j - r_left)
            psi_j = psi_left + slope_2 * (r_j - r_left) + sum_2_max
        else:
            sum_1_left = sum_1_arr[-1]
            sum_2_left = sum_2_arr[-1]
            dr_psi_j = sum_1_left / r_j
            psi_j = sum_1_left * np.log(r_j) - sum_2_left + sum_2_max

        # Calculate fields at r_j.
        if add:
            psi[j] += psi_j
            dr_psi[j] += dr_psi_j
        else:
            psi[j] = psi_j
            dr_psi[j] = dr_psi_j


@njit_serial(fastmath=True, error_model="numpy")
def calculate_psi_and_dr_psi_at_particle_centers(
    r,
    log_r,
    sum_1_arr,
    sum_2_arr,
    psi,
    dr_psi,
):
    # Get number of particles.
    n_part = r.shape[0]

    # Calculate psi after the last plasma plasma particle
    # This is used to ensure the boundary condition psi=0, which also
    # assumes that the total electron and ion charge are the same.
    sum_2_max = sum_2_arr[-1]

    # Calculate fields.
    for i in range(n_part):
        r_i = r[i]
        log_r_i = log_r[i]
        sum_1_i = sum_1_arr[i]
        sum_2_i = sum_2_arr[i]
        dr_psi[i] = sum_1_i / r_i
        psi[i] = sum_1_i * log_r_i - sum_2_i + sum_2_max


@njit_serial()
def calculate_dxi_psi_with_interpolation(r_eval, r, sum_3_arr, dxi_psi, add=False):
    """Calculate dxi_psi at the radial position given in `r_eval`."""
    # Get number of plasma particles.
    n_part = r.shape[0]

    # Get number of points to evaluate.
    n_points = r_eval.shape[0]

    # Calculate dxi_psi after the last plasma plasma particle
    # This is used to ensure the boundary condition dxi_psi=0, which also
    # assumes that the total electron and ion charge are the same.
    sum_3_max = sum_3_arr[-1]

    # Calculate fields at r_eval.
    i_last = 0
    r_left = 0.0
    dxi_psi_left = 0.0
    for j in range(n_points):
        r_j = r_eval[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        while i_last < n_part:
            r_i = r[i_last]
            if r_i >= r_j:
                break
            i_last += 1
        if i_last < n_part:
            if i_last > 0:
                r_left = r[i_last - 1]
                dxi_psi_left = -sum_3_arr[i_last - 1]
            r_right = r[i_last]
            dxi_psi_right = -sum_3_arr[i_last]
            slope = (dxi_psi_right - dxi_psi_left) / (r_right - r_left)
            dxi_psi_j = dxi_psi_left + slope * (r_j - r_left) + sum_3_max
        else:
            dxi_psi_j = -sum_3_arr[-1] + sum_3_max
        if add:
            dxi_psi[j] += dxi_psi_j
        else:
            dxi_psi[j] = dxi_psi_j


@njit_serial(fastmath=True, error_model="numpy")
def calculate_dxi_psi_at_particle_centers(
    r,
    sum_3_arr,
    dxi_psi,
):
    # Get number of particles.
    n_part = r.shape[0]

    # Calculate dxi_psi after the last plasma plasma particle
    # This is used to ensure the boundary condition dxi_psi=0, which also
    # assumes that the total electron and ion charge are the same.
    sum_3_max = sum_3_arr[-1]

    # Calculate fields.
    for i in range(n_part):
        dxi_psi[i] = -sum_3_arr[i] + sum_3_max


@njit_serial()
def check_psi(psi):
    """Check that the values of psi are within a reasonable range

    This is used to prevent issues at the peak of a blowout wake, for example).
    """
    for i in range(psi.shape[0]):
        psi_i = psi[i]
        if psi_i < -0.99:
            psi[i] = -0.99
        elif psi_i > 0.99:
            psi[i] = 0.99


@njit_serial()
def check_psi_derivative(dxi_psi):
    """Check that the values of dxi_psi are within a reasonable range

    This is used to prevent issues at the peak of a blowout wake, for example).
    """
    for i in range(dxi_psi.shape[0]):
        dxi_psi_i = dxi_psi[i]
        if dxi_psi_i < -3:
            dxi_psi[i] = -3
        elif dxi_psi_i > 3:
            dxi_psi[i] = 3
